from __future__ import annotations

from concurrent.futures import Future, ThreadPoolExecutor
import math
from typing import TYPE_CHECKING, Any, Callable, NamedTuple
import warnings
import dataclasses
from qtpy import QtWidgets as QtW
from qtpy import QtGui, QtCore
from qtpy.QtCore import Qt
import numpy as np
from cmap import Colormap
from superqt import ensure_main_thread, QToggleSwitch

from himena.consts import StandardType
from himena.standards import roi, model_meta
from himena.qt._utils import drag_command, qsignal_blocker
from himena.types import DropResult, Parametric, Size, WidgetDataModel
from himena.plugins import validate_protocol, register_function
from himena.widgets import set_status_tip
from himena.data_wrappers import ArrayWrapper, wrap_array
from himena_builtins.qt.widgets._image_components import (
    QImageGraphicsView,
    QRoi,
    QRoiButtons,
    QImageViewControl,
    QImageLabelViewControl,
    QImageViewControlBase,
    QRoiCollection,
    from_standard_roi,
    MouseMode,
)
from himena_builtins.qt.widgets._dim_sliders import QDimsSlider
from himena_builtins.qt.widgets._splitter import QSplitterHandle
from himena_builtins.qt.widgets._shared import quick_min_max

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from himena.style import Theme


class QImageViewBase(QtW.QSplitter):
    _executor = ThreadPoolExecutor(max_workers=1)
    images_changed = QtCore.Signal(list)
    current_roi_updated = QtCore.Signal(object)
    """Emit current QRoi or None if no ROI is active."""

    def __init__(self):
        super().__init__(QtCore.Qt.Orientation.Horizontal)

        widget_left = QtW.QWidget()
        self.addWidget(widget_left)
        layout = QtW.QVBoxLayout(widget_left)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(2)
        self._stick_grid_switch = QToggleSwitch()
        self._stick_grid_switch.setText("Stick to Grid")
        self._stick_grid_switch.setChecked(True)
        self._stick_grid_switch.setToolTip(
            "If checked, vertices of Rectangle and Ellipse ROIs will stick to the\n"
            "grid of the image."
        )
        self._img_view = QImageGraphicsView()
        self._roi_buttons = QRoiButtons(self)
        self._img_view.set_stick_to_grid(True)
        self._img_view.hovered.connect(self._on_hovered)
        self._img_view.roi_visibility_changed.connect(self._roi_visibility_changed)
        self._stick_grid_switch.toggled.connect(self._img_view.set_stick_to_grid)
        self._dims_slider = QDimsSlider()
        self._roi_col = QRoiCollection(self)
        self._roi_col.roi_update_requested.connect(self._update_rois)
        self._roi_col.layout().insertWidget(0, self._roi_buttons)
        self._roi_col.layout().insertWidget(1, self._stick_grid_switch)
        layout.addWidget(self._img_view)
        layout.addWidget(self._dims_slider)

        self._img_view.roi_added.connect(self._on_roi_added)
        self._img_view.roi_removed.connect(self._on_roi_removed)
        self._img_view.current_roi_updated.connect(self._emit_current_roi)
        self._dims_slider.valueChanged.connect(self._slider_changed)

        self.addWidget(self._roi_col)
        self.setStretchFactor(0, 6)
        self.setStretchFactor(1, 1)
        self.setSizes([400, 0])
        self._control = self._make_control_widget()
        self._arr: ArrayWrapper | None = None  # the internal array data for display
        self._is_modified = False  # whether the widget is modified
        self._is_editable = True
        # cached ImageTuples for display
        self._current_image_slices: list[ImageTuple] | None = None
        self._is_rgb = False  # whether the image is RGB
        self._channel_axis: int | None = None
        self._channels: list[ChannelInfo] = [ChannelInfo(name="")]
        self._model_type: str = StandardType.IMAGE
        self._pixel_unit: str = "a.u."
        self._extension_default: str = ".png"
        self._original_title: str | None = None
        self._last_slice_future: Future | None = None
        self._cfg = ImageViewConfigs()

    def createHandle(self):
        return QSplitterHandle(self)

    @property
    def dims_slider(self) -> QDimsSlider:
        """Return the dimension slider widget."""
        return self._dims_slider

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        self._model_type = model.type
        arr = wrap_array(model.value)
        self._original_title = model.title
        is_initialized = self._arr is not None
        is_same_dimensionality = is_initialized and arr.ndim == self._arr.ndim
        is_same_array = is_initialized and (self._arr.arr is model.value)
        ndim_rem = arr.ndim - 2

        # override widget state if metadata is available
        meta0 = model_meta.ImageMeta(
            axes=arr.infer_axes(),
            interpolation="nearest",
            is_rgb=False,
            unit="a.u.",
        )
        self._arr = arr
        if isinstance(meta := model.metadata, model_meta.ImageMeta):
            _update_meta(meta0, meta)
            if meta.is_rgb:
                ndim_rem -= 1

        self._is_rgb = meta0.is_rgb
        if is_initialized:
            sl_old = self._dims_slider.value()
        else:
            sl_old = None
        sl_0 = self._calc_current_indices(arr, meta0, is_same_dimensionality)
        is_sl_same = sl_0 == sl_old

        # update sliders
        if not is_same_array:
            self._dims_slider.set_dimensions(arr.shape, meta0.axes, is_rgb=self._is_rgb)
        with qsignal_blocker(self._dims_slider):
            self._dims_slider.setValue(sl_0)
        axis_names = [meta0.axes[i].name for i in range(self._dims_slider.count())]
        self._roi_col._qroi_list = self._roi_col._qroi_list.coerce_dimensions(
            axis_names
        )

        # update scale bar
        if (axes := meta0.axes) is not None:
            xaxis = axes[-2] if self._is_rgb else axes[-1]
            self._img_view._scale_bar_widget.update_scale_bar(
                scale=xaxis.scale, unit=xaxis.unit
            )

        # update channel info
        if meta0.channel_axis is None or meta0.is_rgb:
            nchannels = 1
        else:
            nchannels = arr.shape[meta0.channel_axis]

        self._img_view.set_n_images(nchannels)
        if meta0.is_rgb:
            self._channel_axis = None  # override channel axis for RGB images
        else:
            self._channel_axis = meta0.channel_axis

        with qsignal_blocker(self._control):
            self._control.update_rgb_channel_dtype(
                is_rgb=self._is_rgb,
                nchannels=nchannels,
                dtype=arr.dtype,
            )

        self._init_channels(meta0, nchannels, arr.dtype)

        if is_same_array and is_sl_same and self._current_image_slices is not None:
            img_slices = self._current_image_slices
        else:
            img_slices = self._get_image_slices(sl_0, nchannels)
        self._update_channels(meta0, img_slices, nchannels)
        if not meta0.skip_image_rerendering:
            self._set_image_slices(img_slices)
        roi_list = meta0.unwrap_rois()
        if len(roi_list) > 0:
            self._roi_col.clear()
            self._roi_col.extend_from_standard_roi_list(
                roi_list.coerce_dimensions(axis_names)
            )
            self._update_rois()
        if cur_roi := meta0.current_roi:
            self._img_view.remove_current_item(reason="update_model")
            if meta0.current_roi_index is None:
                qroi = from_standard_roi(cur_roi, self._img_view._roi_pen)
                self._img_view.set_current_roi(qroi)
            else:
                qroi = self._roi_col._qroi_list[meta0.current_roi_index]
                self._img_view.select_item(qroi, is_registered_roi=True)
            self._img_view.set_mode(MouseMode.from_roi(qroi))
            self._img_view._selection_handles.connect_roi(qroi)

        self._control._interp_check_box.setChecked(meta0.interpolation == "linear")
        self._pixel_unit = meta0.unit or ""
        if ext_default := model.extension_default:
            self._extension_default = ext_default
        self.set_hover_info(self._default_hover_info())
        return None

    @validate_protocol
    def update_value(self, val):
        arr = wrap_array(val)
        if arr.shape != self._arr.shape:
            raise ValueError(
                "`QImageView.update_value` does not support updating the array shape. "
                "Please use `update_model` with `WidgetDataModel` instead."
            )
        self._arr = arr
        self._reset_image()
        self.set_hover_info(self._default_hover_info())

    def _get_image_slices(
        self,
        value: tuple[int, ...],
        nchannels: int,
    ) -> list[ImageTuple]:
        """Get numpy arrays for each channel (None mean hide the channel)."""
        return [ImageTuple(self._get_image_slice_for_channel(value))]

    def _set_image_slice(self, img: NDArray[np.number], channel: ChannelInfo):
        raise NotImplementedError

    def _set_image_slices(self, imgs: list[ImageTuple]):
        """Set image slices using the channel information.

        This method is only used for updating the entire image slices. Channels must be
        correctly set before calling this method, as it uses the channel information to
        transform the image slices.
        """
        raise NotImplementedError

    def _clim_for_ith_channel(self, img_slices: list[ImageTuple], ith: int):
        ar0, _ = img_slices[ith]
        return quick_min_max(ar0)

    def _init_channels(self, meta: model_meta.ImageMeta, nchannels: int, dtype):
        if len(meta.channels) != nchannels:
            ch0 = meta.channels[0]
            channels = [
                ch0.model_copy(update={"colormap": None, "contrast_limits": (0, 1)})
                for _ in range(nchannels)
            ]
        else:
            channels = [
                ch.model_copy(update={"contrast_limits": (0, 1)})
                for ch in meta.channels
            ]
        names = _channel_names(meta, nchannels)
        self._channels = [
            ChannelInfo.from_channel(i, names[i], c) for i, c in enumerate(channels)
        ]

    def _emit_current_roi(self):
        self.current_roi_updated.emit(self._img_view._current_roi_item)
        try:
            idx = self._img_view._roi_items.index(self._img_view._current_roi_item)
        except ValueError:
            pass
        else:
            indices = self._dims_slider.value()
            idx_total = self._roi_col.index_in_slice(indices, idx)
            self._roi_col.set_selections([idx_total])

    def _make_control_widget(self) -> QImageViewControlBase:
        raise NotImplementedError

    def _default_colormap(self) -> Colormap:
        return Colormap("gray")

    def _update_channels(
        self,
        meta: model_meta.ImageMeta,
        img_slices: list[ImageTuple],
        nchannels: int,
    ):
        # before calling ChannelInfo.from_channel, contrast_limits must be set
        if len(meta.channels) != nchannels:
            ch0 = meta.channels[0]
            ch0.contrast_limits = self._clim_for_ith_channel(img_slices, 0)
            channels = [
                ch0.model_copy(update={"colormap": None}) for _ in range(nchannels)
            ]
        else:
            channels = meta.channels
            for i, ch in enumerate(channels):
                if ch.contrast_limits is None:
                    ch.contrast_limits = self._clim_for_ith_channel(img_slices, i)
        names = _channel_names(meta, nchannels)
        self._channels = [
            ChannelInfo.from_channel(i, names[i], c) for i, c in enumerate(channels)
        ]
        if len(self._channels) == 1 and channels[0].colormap is None:
            # ChannelInfo.from_channel returns a single green colormap but it should
            # be gray for single channel images.
            self._channels[0].colormap = self._default_colormap()

    def _calc_current_indices(
        self,
        arr_new: ArrayWrapper,
        meta0: model_meta.ImageMeta,
        is_same_dimensionality: bool,
    ):
        ndim_rem = arr_new.ndim - 3 if meta0.is_rgb else arr_new.ndim - 2
        if meta0.current_indices is not None:
            sl_0 = tuple(force_int(_i) for _i in meta0.current_indices[:ndim_rem])
        elif is_same_dimensionality:
            sl_0 = self._dims_slider.value()
        else:
            if meta0.axes:
                axis_names = [axis.name for axis in meta0.axes][:ndim_rem]
                sl_0 = tuple(
                    size // 2 if aname.lower() == "z" else 0
                    for aname, size in zip(axis_names, arr_new.shape)
                )
            else:
                sl_0 = (0,) * ndim_rem
        # the indices should be in the valid range
        return tuple(
            min(max(0, s), size - 1) for s, size in zip(sl_0, arr_new.shape[:ndim_rem])
        )

    @ensure_main_thread
    def _set_image_slices_async(self, future: Future[list[ImageTuple]]):
        self._last_slice_future = None
        if future.cancelled():
            return
        imgs = future.result()
        self._set_image_slices(imgs)

    def current_roi(self) -> roi.RoiModel | None:
        """Return the current ROI as a standard ROI model."""
        if item := self._img_view._current_roi_item:
            current_roi = item.toRoi()
        else:
            current_roi = None
        return current_roi

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        assert self._arr is not None

        if self._control._interp_check_box.isChecked():
            interp = "linear"
        else:
            interp = "nearest"
        channels = [
            model_meta.ImageChannel(contrast_limits=ch.clim, colormap=ch.colormap.name)
            for ch in self._channels
        ]
        current_indices = self._dims_slider.value()
        current_slices = current_indices + (None, None)
        axes = self._dims_slider.to_dim_axes()
        if self._is_rgb:
            axes.append(model_meta.DimAxis(name="RGB"))
        current_roi = self.current_roi()
        if (
            current_roi is not None
            and not self._img_view._is_current_roi_item_not_registered
            and (sels := self._roi_col.selections())
        ):
            # is the current ROI is already registered, use the index (see the
            # ImageMeta standard)
            current_roi_index = sels[-1]
        else:
            current_roi_index = None
        return WidgetDataModel(
            value=self._arr.arr,
            type=self.model_type(),
            extension_default=self._extension_default,
            metadata=model_meta.ImageMeta(
                current_indices=current_slices,
                axes=axes,
                channels=channels,
                channel_axis=self._channel_axis,
                current_roi=current_roi,
                current_roi_index=current_roi_index,
                rois=self._roi_col.to_standard_roi_list,
                is_rgb=self._is_rgb,
                interpolation=interp,
                unit=self._pixel_unit,
            ),
        )

    @validate_protocol
    def model_type(self) -> str:
        return self._model_type

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 400, 400

    @validate_protocol
    def is_editable(self) -> bool:
        return self._is_editable

    @validate_protocol
    def set_editable(self, editable: bool):
        self._is_editable = editable

    @validate_protocol
    def control_widget(self) -> QImageViewControl:
        return self._control

    @validate_protocol
    def allowed_drop_types(self) -> list[str]:
        return [StandardType.ROIS, StandardType.IMAGE_LABELS]

    @validate_protocol
    def dropped_callback(self, model: WidgetDataModel):
        if model.type == StandardType.ROIS:
            # dropping ROIs to concatenate to the current ROI list
            if isinstance(roi_list := model.value, roi.RoiListModel):
                self._roi_col.extend_from_standard_roi_list(roi_list)
                self._update_rois()
            self._is_modified = True
            return DropResult(delete_input=False)
        elif model.type == StandardType.IMAGE_LABELS:
            raise NotImplementedError("Merging with labels is not implemented yet.")
        return None

    @validate_protocol
    def widget_resized_callback(self, old: Size, new: Size):
        self._img_view.resize_event(old, new)

    @validate_protocol
    def widget_added_callback(self):
        self._img_view.auto_range()
        self._img_view.update_handle_sizes()

    def setFocus(self):
        return self._img_view.setFocus()

    def leaveEvent(self, ev) -> None:
        self.set_hover_info(self._default_hover_info())

    def _roi_visibility_changed(self, show_rois: bool):
        with qsignal_blocker(self._roi_col):
            self._roi_col._roi_visible_btn.setChecked(show_rois)

    def _update_image_visibility(self, visible: list[bool]):
        slices = self._current_image_slices
        if slices is None:
            return
        for i, vis in enumerate(visible):
            if i < len(slices):
                slices[i] = ImageTuple(slices[i].arr, vis)
        self._set_image_slices(slices)

    def _get_image_slice_for_channel(
        self, value: tuple[int, ...]
    ) -> NDArray[np.number]:
        """Get numpy array for current channel."""
        return self._arr.get_slice(tuple(value))

    def _slider_changed(self, value: tuple[int, ...], *, force_sync: bool = False):
        """Callback for slider value change."""
        if self._arr is None:
            return
        if self._last_slice_future:
            self._last_slice_future.cancel()  # cancel last task
        if force_sync:
            img_slices = self._get_image_slices(value, len(self._channels))
            self._set_image_slices(img_slices)
        else:
            # set slice asynchronously
            self._last_slice_future = self._executor.submit(
                self._get_image_slices, value, len(self._channels)
            )
            self._last_slice_future.add_done_callback(self._set_image_slices_async)

    def _update_rois(self):
        cur_item = self._img_view._current_roi_item
        self._img_view.clear_rois()
        rois = self._roi_col.get_rois_on_slice(self._dims_slider.value())
        self._img_view.extend_qrois(rois, cur_item)

    def current_channel(self, slider_value: tuple[int] | None = None) -> ChannelInfo:
        """Get the current channel info.

        This method always returns a channel, as even a monochrome image is set with a
        gray colormap.
        """
        if slider_value is None:
            _slider_value = self._dims_slider.value()
        else:
            _slider_value = slider_value
        if self._channel_axis is not None:
            ith = _slider_value[self._channel_axis]
            ch = self._channels[ith]
        else:
            ch = self._channels[0]
        return ch

    def _on_roi_added(self, qroi: QRoi):
        if qroi.label() == "":
            qroi.set_label(roi.default_roi_label(len(self._roi_col)))
        indices = self._dims_slider.value()
        self._roi_col.add(indices, qroi)
        set_status_tip(f"Added a {qroi._roi_type()} ROI")

    def _on_roi_removed(self, idx: int):
        indices = self._dims_slider.value()
        qroi = self._roi_col.pop_roi_in_slice(indices, idx)
        set_status_tip(f"Removed a {qroi._roi_type()} ROI")
        self._roi_col.set_selections([])  # deselect

    def _reset_image(self):
        if self._channels is None:  # not initialized yet
            return
        imgs = self._get_image_slices(self._dims_slider.value(), len(self._channels))
        self._set_image_slices(imgs)

    def _on_hovered(self, pos: QtCore.QPointF):
        """Update hover info by the pixel position and value."""
        x, y = pos.x(), pos.y()
        if self._current_image_slices is None:
            return
        iy, ix = int(y), int(x)
        idx = self.current_channel().channel_index or 0
        cur_img = self._current_image_slices[idx]
        ny, nx, *_ = cur_img.arr.shape
        if 0 <= iy < ny and 0 <= ix < nx:
            if not cur_img.visible:
                self.set_hover_info(f"x={x:.1f}, y={y:.1f} (invisible)")
                return
            intensity = cur_img.arr[int(y), int(x)]
            # NOTE: `intensity` could be an RGBA numpy array.
            if isinstance(intensity, np.ndarray) or self._arr.dtype.kind == "b":
                _int = str(intensity)
            else:
                if self._arr.dtype.kind == "f":
                    fmt = ".3g"
                else:
                    fmt = ".0f"
                _int = format(intensity, fmt)
            if self._pixel_unit:
                _int += f" [{self._pixel_unit}]"
            yaxis, xaxis = self.dims_slider._yx_axes
            if self._model_type == StandardType.IMAGE_FOURIER:
                kx = (ix - nx // 2) / nx
                ky = (iy - ny // 2) / ny
                xinfo = _fourier_info(kx, xaxis)
                yinfo = _fourier_info(ky, yaxis)
                angle = math.degrees(math.atan2(ky, kx))
                self.set_hover_info(
                    f"kx={xinfo}, ky={yinfo}, theta={angle:.1f} value={_int}"
                )
            else:
                xinfo = _physical_info(ix, xaxis)
                yinfo = _physical_info(iy, yaxis)
                self.set_hover_info(f"x={xinfo}, y={yinfo}, value={_int}")
        else:
            self.set_hover_info(self._default_hover_info())

    def set_hover_info(self, text: str):
        """Set the hover info text to the control widget."""
        self._control._hover_info.setText(text)

    def _default_hover_info(self) -> str:
        if self._arr is None:
            return
        return f"{self._arr.shape}, {self._arr.dtype}"

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        if event is None or event.isAutoRepeat():
            return None
        _mods = event.modifiers()
        _key = event.key()
        view = self._img_view
        if _mods & Qt.KeyboardModifier.ControlModifier:
            view.standard_ctrl_key_press(_key)
        elif event.text() == "^":
            self.handle(0).toggle()
        else:
            view.standard_key_press(_key, _mods & Qt.KeyboardModifier.ShiftModifier)
        return None

    def keyReleaseEvent(self, event: QtGui.QKeyEvent | None) -> None:
        if event is None or event.isAutoRepeat():
            return None
        if event.key() == Qt.Key.Key_Space:
            self._img_view.set_mode(self._img_view._last_mode_before_key_hold)
        return None

    def _on_drag_roi_requested(self, indices: list[int]):
        nrois = len(indices)
        _s = "" if nrois == 1 else "s"
        return drag_command(
            self,
            command_id=_COMMAND_ID,
            type=StandardType.ROIS,
            with_params={"selections": indices},
            desc=f"{nrois} ROI{_s}",
        )


_COMMAND_ID = "builtins:select-image-rois"


@register_function(menus=[], command_id=_COMMAND_ID)
def _select_image_rois(model: WidgetDataModel) -> Parametric:
    assert isinstance(meta := model.metadata, model_meta.ImageMeta)
    rois = meta.unwrap_rois()
    axes = meta.axes

    def run_select(selections: list[int]) -> WidgetDataModel:
        if len(selections) == 0:
            raise ValueError("No ROIs selected.")

        return WidgetDataModel(
            value=rois.filter_by_selection(selections),
            type=StandardType.ROIS,
            title=f"Subset of {model.title}",
            metadata=model_meta.ImageRoisMeta(axes=axes),
        )

    return run_select


class QImageView(QImageViewBase):
    """The built-in n-D image viewer widget.

    ## Basic Usage

    The image canvas can be interactively panned by dragging with the left mouse button,
    and zoomed by scrolling the mouse wheel. Zooming is reset by double-click. This
    widget also supports adding ROIs (Regions of Interest) to the image.

    - For multi-dimensional images, dimension sliders will be added to the bottom of the
      image canvas.
    - For multi-channel images, channel composition mode can be selected in the control
      widget. Toggle switches for each channel are available next to it.
    - The histogram of the current channel is displayed in the control widget. The
      contrast limits can be adjusted by dragging the handles on the histogram.
    - A ROI manager is in the right collapsible panel. Press the ROI buttons to switch
      between ROI modes.

    ## Keyboard Shortcuts

    - `L`: Switch to Line ROI.
    - `Shift+L`: Switch to Segmented Line ROI.
    - `R`: Switch to Rectangle ROI.
    - `Shift+R`: Switch to Rotated Rectangle ROI.
    - `E`: Switch to Ellipse ROI.
    - `Shift+E`: Switch to Rotated Ellipse ROI.
    - `C`: Switch to Circle ROI.
    - `P`: Switch to Point ROI.
    - `Shift+P`: Switch to Multi-Point ROI.
    - `G`: Switch to Polygon ROI.
    - `Z`: Switch to Pan/Zoom Mode.
    - `Space`: Hold this key to temporarily switch to Pan/Zoom Mode.
    - `S`: Switch to Select Mode.
    - `T`: Add current ROI to the ROI list.
    - `V`: Toggle visibility of all the added ROIs.
    - `Delete` / `Backspace`: Remove the current ROI.
    - `Ctrl+A`: Select the entire image with a rectangle ROI.
    - `Ctrl+X`: Cut the selected ROI.
    - `Ctrl+C`: Copy the selected ROI.
    - `Ctrl+V`: Paste the copied ROI.
    - `Ctrl+D`: Duplicate the selected ROI.

    ## Drag and Drop

    - This widget accepts dropping models with `StandardType.ROIS` ("rois").
      Dropped ROIs will be added to the ROI list.
    - ROIs can be dragged out from the ROI manager. The type of the dragged model is
      `StandardType.ROIS` ("rois"). Use the drag indicator in the corner of the ROI
      list.
    """

    __himena_widget_id__ = "builtins:QImageView"
    __himena_display_name__ = "Built-in Image Viewer"
    _control: QImageViewControl

    @validate_protocol
    def theme_changed_callback(self, theme: Theme):
        if theme.name.startswith("light"):
            color = QtGui.QColor(0, 0, 0)
        else:
            color = QtGui.QColor(255, 255, 255)
        self._roi_buttons._update_colors(color)

    @validate_protocol
    def update_configs(self, cfg: ImageViewConfigs):
        self._cfg = cfg
        self._img_view.set_stick_to_grid(cfg.default_stick_to_grid)
        self._img_view.set_show_rois(cfg.default_show_all)
        self._img_view.set_show_labels(cfg.default_show_labels)
        self._img_view._selection_handles._handle_size = cfg.roi_handle_size

    def _make_control_widget(self) -> QImageViewControl:
        return QImageViewControl(self)

    def _update_channels(self, meta, img_slices, nchannels):
        super()._update_channels(meta, img_slices, nchannels)
        if self._composite_state() == "Comp.":
            self._control._chn_vis.set_channels(self._channels)

    def _composite_state(self) -> str:
        return self._control._chn_mode_combo.currentText()

    def _get_image_slices(
        self,
        value: tuple[int, ...],
        nchannels: int,
    ) -> list[ImageTuple]:
        """Get numpy arrays for each channel (None mean hide the channel)."""
        if self._channel_axis is None:
            return super()._get_image_slices(value, nchannels)

        img_slices = []
        check_states = self._control._channel_visibility()
        for i in range(nchannels):
            vis = i >= len(check_states) or check_states[i]
            sl = list(value)
            sl[self._channel_axis] = i
            img_slices.append(ImageTuple(self._get_image_slice_for_channel(sl), vis))
        return img_slices

    def _set_image_slice(self, img: NDArray[np.number], channel: ChannelInfo):
        idx = channel.channel_index or 0
        with qsignal_blocker(self._control._histogram):
            self._img_view.set_array(
                idx,
                channel.transform_image(
                    img,
                    complex_transform=self._control.complex_transform,
                    is_rgb=self._is_rgb,
                    is_gray=self._composite_state() == "Gray",
                ),
            )
            self._img_view.clear_rois()
            self._control._histogram.set_hist_for_array(
                img,
                clim=channel.clim,
                is_rgb=self._is_rgb,
                color=color_for_colormap(channel.colormap),
            )
        if self._current_image_slices is not None:
            visible = self._current_image_slices[idx].visible
        else:
            visible = True
        self._current_image_slices[idx] = ImageTuple(img, visible)
        self.images_changed.emit(self._current_image_slices)

    def _set_image_slices(self, imgs: list[ImageTuple]):
        """Set image slices using the channel information.

        This method is only used for updating the entire image slices. Channels must be
        correctly set before calling this method, as it uses the channel information to
        transform the image slices.
        """
        self._current_image_slices = imgs
        if self._channels is None:
            return
        images: list[NDArray[np.number] | None] = []
        with qsignal_blocker(self._control._histogram):
            for i, (imtup, ch) in enumerate(zip(imgs, self._channels)):
                if imtup.visible:
                    img = imtup.arr
                else:
                    img = None
                images.append(img)
                self._img_view.set_array(
                    i,
                    ch.transform_image(
                        img,
                        complex_transform=self._control.complex_transform,
                        is_rgb=self._is_rgb,
                        is_gray=self._composite_state() == "Gray",
                    ),
                )
            ch_cur = self.current_channel()
            idx = ch_cur.channel_index or 0
            hist_arr_ref = imgs[idx].arr
            if hist_arr_ref.dtype.kind == "c":
                hist_arr_ref = self._control.complex_transform(hist_arr_ref)
            self._control._histogram.set_hist_for_array(
                hist_arr_ref,
                clim=ch_cur.clim,
                is_rgb=self._is_rgb,
                color=color_for_colormap(ch_cur.colormap),
            )
            self._update_rois()
            self._img_view.set_image_blending([im.visible for im in imgs])
        self.images_changed.emit(images)

    def _clim_for_ith_channel(self, img_slices: list[ImageTuple], ith: int):
        ar0, _ = img_slices[ith]
        if ar0.dtype.kind == "c":
            ar0 = self._control.complex_transform(ar0)
        return quick_min_max(ar0)


class QImageLabelView(QImageViewBase):
    """The default nD image label viewer widget.

    Unlink the image viewer, this widget displays image labels by discrete colors.

    ## Basic Usage

    The image canvas can be interactively panned by dragging with the left mouse button,
    and zoomed by scrolling the mouse wheel. Zooming is reset by double-click. This
    widget also supports adding ROIs (Regions of Interest) to the image.

    - For multi-dimensional images, dimension sliders will be added to the bottom of the
      image canvas.
    - For multi-channel images, channel composition mode can be selected in the control
      widget. Toggle switches for each channel are available next to it.
    - The histogram of the current channel is displayed in the control widget. The
      contrast limits can be adjusted by dragging the handles on the histogram.
    - A ROI manager is in the right collapsible panel. Press the ROI buttons to switch
      between ROI modes.

    ## Keyboard Shortcuts

    - `L`: Switch to Line ROI.
    - `Shift+L`: Switch to Segmented Line ROI.
    - `R`: Switch to Rectangle ROI.
    - `Shift+R`: Switch to Rotated Rectangle ROI.
    - `E`: Switch to Ellipse ROI.
    - `Shift+E`: Switch to Rotated Ellipse ROI.
    - `C`: Switch to Circle ROI.
    - `P`: Switch to Point ROI.
    - `Shift+P`: Switch to Multi-Point ROI.
    - `G`: Switch to Polygon ROI.
    - `Z`: Switch to Pan/Zoom Mode.
    - `Space`: Hold this key to temporarily switch to Pan/Zoom Mode.
    - `S`: Switch to Select Mode.
    - `T`: Add current ROI to the ROI list.
    - `V`: Toggle visibility of all the added ROIs.
    - `Delete` / `Backspace`: Remove the current ROI.
    - `Ctrl+A`: Select the entire image with a rectangle ROI.
    - `Ctrl+X`: Cut the selected ROI.
    - `Ctrl+C`: Copy the selected ROI.
    - `Ctrl+V`: Paste the copied ROI.
    - `Ctrl+D`: Duplicate the selected ROI.

    ## Drag and Drop

    - This widget accepts dropping models with `StandardType.IMAGE_ROIS` ("image-rois").
      Dropped ROIs will be added to the ROI list.
    - ROIs can be dragged out from the ROI manager. The type of the dragged model is
      `StandardType.IMAGE_ROIS` ("image-rois"). Use the drag indicator in the corner of
      the ROI list.
    """

    __himena_widget_id__ = "builtins:QImageLabelView"
    __himena_display_name__ = "Built-in Image Label Viewer"
    _control: QImageLabelViewControl

    def _default_colormap(self) -> Colormap:
        return Colormap("gnuplot:rainbow")

    def _make_control_widget(self) -> QImageLabelViewControl:
        return QImageLabelViewControl(self)

    def _get_opacity(self) -> float:
        return self._control._opacity_slider.value()

    def _set_image_slice(self, img: NDArray[np.number], channel: ChannelInfo):
        idx = channel.channel_index or 0
        self._img_view.set_array(
            idx, channel.transform_labels(img, opacity=self._get_opacity())
        )
        self._img_view.clear_rois()
        self._current_image_slices[idx] = ImageTuple(img)

    def _set_image_slices(self, imgs: list[ImageTuple]):
        """Set image slices using the channel information.

        This method is only used for updating the entire image slices. Channels must be
        correctly set before calling this method, as it uses the channel information to
        transform the image slices.
        """
        self._current_image_slices = imgs
        if self._channels is None:
            return
        for i, (imtup, ch) in enumerate(zip(imgs, self._channels)):
            self._img_view.set_array(
                i, ch.transform_labels(imtup.arr, opacity=self._get_opacity())
            )
        self._update_rois()
        self._img_view.set_image_blending([im.visible for im in imgs])


@dataclasses.dataclass
class ChannelInfo:
    name: str
    clim: tuple[float, float] = dataclasses.field(default=(0.0, 1.0))
    colormap: Colormap = dataclasses.field(default_factory=lambda: Colormap("gray"))
    channel_index: int | None = None

    def transform_image(
        self,
        arr: NDArray[np.number] | None,
        complex_transform: Callable[
            [NDArray[np.complexfloating]], NDArray[np.number]
        ] = np.abs,
        is_rgb: bool = False,
        is_gray: bool = False,
    ) -> NDArray[np.uint8] | None:
        """Convenience method to transform the array to a displayable RGBA image."""
        if is_rgb:
            return self.transform_image_rgb(arr, is_gray=is_gray)
        else:
            if is_gray:
                return self.as_gray().transform_image(arr, complex_transform)
            return self.transform_image_2d(arr, complex_transform)

    def transform_labels(
        self,
        arr: NDArray[np.uint] | None,
        opacity: float = 0.7,
        period: int = 256,
    ) -> NDArray[np.uint8] | None:
        """Transform an uint array to a displayable RGBA image."""
        if arr is None:
            return None
        mask_background = arr == 0
        if arr.dtype == np.uint8:
            # NOTE: arr % 256 raises OverflowError for arr.dtype == np.uint8.
            arr_index = arr.copy()
        else:
            arr_index = np.empty(arr.shape, dtype=np.uint8)
            np.mod(arr - 1, period, out=arr_index)
        arr_index *= 109
        arr_normed = self.colormap(arr_index, N=period, bytes=True)
        arr_normed[:, :, 3] = int(255 * opacity)
        arr_normed[mask_background] = 0
        out = np.ascontiguousarray(arr_normed)
        return out

    def transform_image_2d(
        self,
        arr: NDArray[np.number] | None,
        complex_transform: Callable[[NDArray[np.complexfloating]], NDArray[np.number]],
    ) -> NDArray[np.uint8] | None:
        """Transform the array to a displayable RGBA image."""
        if arr is None:
            return None
        if arr.ndim == 3:
            return arr  # RGB
        cmin, cmax = self.clim
        if arr.dtype.kind == "c":
            arr = complex_transform(arr)
        if arr.dtype.kind == "b":
            false_color = (np.array(self.colormap(0.0).rgba) * 255).astype(np.uint8)
            true_color = (np.array(self.colormap(1.0).rgba) * 255).astype(np.uint8)
            arr_normed = np.where(arr[..., np.newaxis], true_color, false_color)
        elif cmax > cmin:
            arr = arr.clip(cmin, cmax)
            arr_normed = (self.colormap((arr - cmin) / (cmax - cmin)) * 255).astype(
                np.uint8
            )
        else:
            color = np.array(self.colormap(0.5).rgba8, dtype=np.uint8)
            arr_normed = np.empty(arr.shape + (4,), dtype=np.uint8)
            arr_normed[:] = color[np.newaxis, np.newaxis]
        out = np.ascontiguousarray(arr_normed)
        assert out.dtype == np.uint8
        return out

    def transform_image_rgb(
        self,
        arr: NDArray[np.number] | None,
        is_gray: bool = False,
    ):
        """Transform the RGBA array to a displayable RGBA image."""
        if arr is None:
            return None
        cmin, cmax = self.clim
        if cmax == cmin:
            amp = 128
        else:
            amp = 255 / (cmax - cmin)
        if is_gray:
            # make a gray image
            arr_gray = arr[..., 0] * 0.3 + arr[..., 1] * 0.59 + arr[..., 2] * 0.11
            arr_gray = arr_gray.astype(np.uint8)
            if arr.shape[2] == 4:
                alpha = arr[..., 3]
            else:
                alpha = np.full(arr_gray.shape, 255, dtype=np.uint8)
            arr = np.stack([arr_gray, arr_gray, arr_gray, alpha], axis=-1)
        if (cmin, cmax) == (0, 255):
            arr_normed = arr
        else:
            if arr.shape[2] == 3:
                arr_normed = ((arr - cmin) * amp).clip(0, 255).astype(np.uint8)
            else:
                arr_normed = arr.copy()
                if is_gray:
                    sl = slice(None)
                else:
                    sl = (slice(None), slice(None), slice(None, 3))
                arr_normed[sl] = ((arr[sl] - cmin) * amp).clip(0, 255).astype(np.uint8)
        return arr_normed

    def as_gray(self) -> ChannelInfo:
        return ChannelInfo(
            name=self.name,
            clim=self.clim,
            colormap=Colormap("gray"),
            channel_index=self.channel_index,
        )

    @classmethod
    def from_channel(
        cls,
        idx: int,
        name: str,
        channel: model_meta.ImageChannel,
    ) -> ChannelInfo:
        input_clim = channel.contrast_limits
        if input_clim is None:
            raise ValueError("Contrast limits are not set.")
        if channel.colormap is None:
            colormap = Colormap(f"cmap:{_DEFAULT_COLORMAPS[idx % 6]}")
        else:
            colormap = Colormap(channel.colormap)
        return cls(
            name=name,
            channel_index=idx,
            colormap=colormap,
            clim=channel.contrast_limits,
        )


class ImageTuple(NamedTuple):
    """A layer of image and its visibility."""

    arr: NDArray[np.number]
    visible: bool = True


_DEFAULT_COLORMAPS = ["green", "magenta", "cyan", "yellow", "red", "blue"]


def color_for_colormap(cmap: Colormap) -> QtGui.QColor:
    """Get the representative color for the colormap."""
    return QtGui.QColor.fromRgbF(*cmap(0.5))


def force_int(idx: Any) -> int:
    if isinstance(idx, int):
        return idx
    if hasattr(idx, "__index__"):
        return idx.__index__()
    warnings.warn(f"Cannot convert {idx} to int, using 0 instead.")
    return 0


def _update_meta(meta0: model_meta.ImageMeta, meta: model_meta.ImageMeta):
    if meta.rois:
        meta0.rois = meta.rois
    if meta.axes:
        meta0.axes = meta.axes
    if meta.is_rgb:
        meta0.is_rgb = meta.is_rgb
        meta0.interpolation = "linear"
    if meta.channel_axis is not None:
        meta0.channel_axis = meta.channel_axis
    if meta.current_indices:
        meta0.current_indices = meta.current_indices
    if meta.channels:
        meta0.channels = meta.channels
    if meta.current_roi:
        meta0.current_roi = meta.current_roi
    if meta.unit is not None:
        meta0.unit = meta.unit
    if meta.interpolation:
        meta0.interpolation = meta.interpolation
    meta0.current_roi_index = meta.current_roi_index
    meta0.skip_image_rerendering = meta.skip_image_rerendering
    return None


def _channel_names(meta: model_meta.ImageMeta, nchannels: int) -> list[str]:
    if meta.channel_axis is None:
        names = [""]
    elif meta.axes is None:
        names = [f"Ch-{i}" for i in range(nchannels)]
    else:
        names = [meta.axes[meta.channel_axis].get_label(i) for i in range(nchannels)]
    return names


def _physical_info(xpix: float, axis: model_meta.DimAxis) -> str:
    xpos = xpix * axis.scale + axis.origin
    digit = _digits_for_scale(axis.scale)
    if axis.unit:
        return f"{int(xpix)} ({xpos:.{digit}f} {axis.unit})"
    else:
        return f"{int(xpix)} ({xpos:.{digit}f})"


def _fourier_info(kx: float, axis: model_meta.DimAxis) -> str:
    if kx == 0:
        wlen = float("inf")
    else:
        wlen = axis.scale / kx
    digit = _digits_for_scale(axis.scale)
    if axis.unit:
        return f"{kx:.2f} ({wlen:.{digit}f} /c)"
    else:
        return f"{kx:.2f} ({wlen:.{digit}f} {axis.unit}/c)"


def _digits_for_scale(scale: float) -> int:
    if scale < 10:
        return -math.floor(math.log10(scale)) + 1
    else:
        return 0


@dataclasses.dataclass
class ImageViewConfigs:
    default_stick_to_grid: bool = dataclasses.field(default=True)
    roi_handle_size: int = dataclasses.field(default=5, metadata={"min": 1, "max": 10})
    default_show_all: bool = dataclasses.field(default=True)
    default_show_labels: bool = dataclasses.field(default=True)



from qtpy.QtWidgets import QApplication
from pytestqt.qtbot import QtBot
from himena import MainWindow
from himena.testing import WidgetTester
from himena_builtins.qt.basic import QWorkflowView
from himena_builtins._io import default_workflow_reader
from qtpy.QtCore import Qt
from pathlib import Path

_Ctrl = Qt.KeyboardModifier.ControlModifier

def test_workflow_view(qtbot: QtBot, sample_dir: Path):
    widget = QWorkflowView()
    widget.show()
    qtbot.addWidget(widget)
    with WidgetTester(widget) as tester:
        tester.update_model(default_workflow_reader(sample_dir / "test.workflow.json"))
        tester.cycle_model()
        QApplication.processEvents()
        vp = widget._tree_widget.viewport()
        qtbot.keyClick(vp, Qt.Key.Key_Down)
        qtbot.keyClick(vp, Qt.Key.Key_Up)
        qtbot.keyClick(vp, Qt.Key.Key_Down, _Ctrl)
        qtbot.keyClick(vp, Qt.Key.Key_Up, _Ctrl)
        widget._tree_widget.update()
        widget._make_context_menu(widget._tree_widget.topLevelItem(0))
        pos = widget._tree_widget.visualItemRect(widget._tree_widget.topLevelItem(0)).center()
        widget._tree_widget._make_drag(pos)

def test_edit_workflow_view(qtbot: QtBot, sample_dir: Path):
    widget = QWorkflowView()
    widget.show()
    qtbot.addWidget(widget)
    with WidgetTester(widget) as tester:
        tester.update_model(default_workflow_reader(sample_dir / "test.workflow.json"))
        widget._toggle_to_be_added(widget._tree_widget.topLevelItem(0))
        widget._replace_with_file_reader(widget._tree_widget.topLevelItem(0), "file")
        widget._replace_with_file_reader(widget._tree_widget.topLevelItem(0), "model")

def test_find_window(himena_ui: MainWindow):
    win0 = himena_ui.add_object("a")
    himena_ui.exec_action("duplicate-window")
    himena_ui.exec_action("show-workflow-graph")
    win1 = himena_ui.current_window
    assert isinstance(win1.widget, QWorkflowView)
    win1.widget._find_window(win1.widget._tree_widget.topLevelItem(0))
    assert himena_ui.current_window is win0



from __future__ import annotations
from typing import Any, TYPE_CHECKING

from magicgui.types import Undefined
from magicgui.widgets import LineEdit, ComboBox
from magicgui.widgets.bases import ValuedContainerWidget, ValueWidget
from cmap import Colormap
from himena.qt.magicgui._color import ColorEdit, ColormapEdit
from himena.qt.magicgui._basic_widgets import FloatEdit
from himena.qt.magicgui._toggle_switch import ToggleSwitch
from himena._utils import to_color_or_colormap
from himena.consts import MonospaceFontFamily

if TYPE_CHECKING:
    from typing import TypedDict, NotRequired

    class FacePropertyDict(TypedDict):
        color: NotRequired[str | dict[str, Any]]
        hatch: NotRequired[str]

    class EdgePropertyDict(TypedDict):
        color: NotRequired[str | dict[str, Any]]
        width: NotRequired[float]
        style: NotRequired[str]

    class AxisPropertyDict(TypedDict):
        lim: NotRequired[tuple[float, float]]
        scale: NotRequired[str]
        label: NotRequired[str]
        ticks: NotRequired[Any]
        grid: NotRequired[bool]


class ColorOrColorCycleEdit(ValuedContainerWidget):
    def __init__(self, value=Undefined, **kwargs):
        self._use_color_cycle = ToggleSwitch(value=False, text="use color cycle")
        self._color = ColorEdit(value="black")
        self._color_cycle = ColormapEdit(
            value="tab10",
            defaults=[
                "tab10",
                "tab10_colorblind",
                "tab20",
                "colorbrewer:Accent",
                "colorbrewer:Dark2",
                "colorbrewer:Paired",
                "colorbrewer:Pastel1",
                "colorbrewer:Pastel2",
                "colorbrewer:Set1",
                "colorbrewer:Set2",
                "colorbrewer:Set3",
            ],
            visible=False,
        )
        super().__init__(
            value,
            widgets=[self._use_color_cycle, self._color, self._color_cycle],
            **kwargs,
        )
        self.margins = (0, 0, 0, 0)
        self.min_height = 60
        self._use_color_cycle.changed.connect(self._update_visibility)

    def _update_visibility(self, v: bool):
        self._color_cycle.visible = v
        self._color.visible = not v

    def get_value(self) -> Any:
        if self._use_color_cycle.value:
            cmap_value = self._color_cycle.value
            if cmap_value.name == "custom colormap":
                value = cmap_value.as_dict()
            else:
                value = cmap_value.name
        else:
            value = self._color.value.hex
        if value is None and not self._nullable:
            raise ValueError("Value cannot be None")
        return value

    def set_value(self, value: Any):
        value = to_color_or_colormap(value)
        is_cycle = isinstance(value, Colormap)
        self._use_color_cycle.value = is_cycle
        if is_cycle:
            self._color_cycle.value = value
        else:
            self._color.value = value


class FacePropertyEdit(ValuedContainerWidget["FacePropertyDict"]):
    def __init__(self, value=Undefined, **kwargs):
        if value is None:
            value = Undefined
        self._face_color = ColorOrColorCycleEdit(value="#FFFFFF", label="color")
        self._face_hatch = LineEdit(value="", label="hatch")
        super().__init__(
            value,
            widgets=[self._face_color, self._face_hatch],
            **kwargs,
        )
        self._face_color.changed.connect(self._emit_value_changed)
        self._face_hatch.changed.connect(self._emit_value_changed)
        self.margins = (0, 0, 0, 0)

    def _emit_value_changed(self):
        self.changed.emit(self.get_value())

    def get_value(self) -> FacePropertyDict:
        return {
            "color": self._face_color.value,
            "hatch": self._face_hatch.value,
        }

    def set_value(self, value: FacePropertyDict):
        value = value or {}
        self._face_color.value = value.get("color", "white")
        self._face_hatch.value = value.get("hatch", "")


class EdgePropertyEdit(ValuedContainerWidget["EdgePropertyDict"]):
    _STYLE_CHOICES = [
        ("———————", "-"),
        ("— — — —", "--"),
        ("-·-·-·-", "-."),
        ("-··-··-", "-.."),
        ("·······", ":"),
    ]

    def __init__(self, value=Undefined, **kwargs):
        if value is None:
            value = Undefined
        self._edge_color = ColorOrColorCycleEdit(value="#000000", label="color")
        self._edge_width = FloatEdit(value=1.0, label="width", min=0.0)
        self._edge_style = ComboBox(
            value="-", choices=self._STYLE_CHOICES, label="style"
        )
        self._edge_style.native.setStyleSheet(f"font-family: {MonospaceFontFamily}")
        super().__init__(
            value,
            widgets=[self._edge_color, self._edge_width, self._edge_style],
            **kwargs,
        )
        self._edge_color.changed.connect(self._emit_value_changed)
        self._edge_width.changed.connect(self._emit_value_changed)
        self._edge_style.changed.connect(self._emit_value_changed)
        self.changed.connect(self._on_property_changed)
        self._on_property_changed(self.get_value())

    def get_value(self) -> EdgePropertyDict:
        return {
            "color": self._edge_color.value,
            "width": round(self._edge_width.value, 2),
            "style": self._edge_style.value,
        }

    def set_value(self, value: EdgePropertyDict):
        value = value or {}
        with self.changed.blocked():
            self._edge_color.value = value.get("color", "black")
            ewidth = value.get("width")
            if ewidth is None:
                ewidth = 1.0
            self._edge_width.value = round(ewidth, 2)
            estyle = value.get("style")
            if estyle is None:
                estyle = "-"
            self._edge_style.value = estyle
        self._emit_value_changed()

    def _emit_value_changed(self) -> None:
        self.changed.emit(self.get_value())

    def _on_property_changed(self, value: EdgePropertyDict):
        if value is None:
            return
        enabled = value["width"] > 0.0
        self._edge_color.enabled = enabled
        self._edge_style.enabled = enabled


class LimitEdit(ValuedContainerWidget[tuple[float, float]]):
    def __init__(self, value=Undefined, **kwargs):
        self._min_widget = FloatEdit(value=0.0, label="min")
        self._max_widget = FloatEdit(value=1.0, label="max")
        super().__init__(
            value,
            widgets=[self._min_widget, self._max_widget],
            layout="horizontal",
            **kwargs,
        )
        self._min_widget.changed.connect(self._emit_value_changed)
        self._max_widget.changed.connect(self._emit_value_changed)
        self.margins = (0, 0, 0, 0)

    def get_value(self) -> tuple[float, float]:
        return self._min_widget.value, self._max_widget.value

    def set_value(self, value: tuple[float, float]):
        with self.changed.blocked():
            self._min_widget.value = value[0]
            self._max_widget.value = value[1]
        self._emit_value_changed()

    def _emit_value_changed(self) -> None:
        self.changed.emit(self.get_value())


class AxisPropertyEdit(ValuedContainerWidget["AxisPropertyDict"]):
    def __init__(self, value=Undefined, **kwargs):
        if value is None:
            value = Undefined
        self._lim_widget = LimitEdit(label="lim")
        self._scale_widget = ComboBox(choices=["linear", "log"], label="scale")
        self._label_widget = LineEdit(value="", label="label")
        self._grid_widget = ToggleSwitch(value=False, text="grid")
        super().__init__(
            value,
            widgets=[
                self._lim_widget,
                self._scale_widget,
                self._label_widget,
                self._grid_widget,
            ],
            **kwargs,
        )
        self._lim_widget.changed.connect(self._emit_value_changed)
        self._scale_widget.changed.connect(self._emit_value_changed)
        self._label_widget.changed.connect(self._emit_value_changed)
        self._grid_widget.changed.connect(self._emit_value_changed)

    def get_value(self) -> AxisPropertyDict:
        return {
            "lim": self._lim_widget.value,
            "scale": self._scale_widget.value,
            "label": self._label_widget.value,
            "grid": self._grid_widget.value,
        }

    def set_value(self, value: AxisPropertyDict):
        value = value or {}
        with self.changed.blocked():
            if (lim := value.get("lim", None)) is not None:
                self._lim_widget.value = tuple(lim)
            if "scale" in value:
                self._scale_widget.value = value["scale"]
            if (label := value.get("label", None)) is not None:
                self._label_widget.value = label
            if "grid" in value:
                self._grid_widget.value = value["grid"]
        self._emit_value_changed()

    def _emit_value_changed(self) -> None:
        self.changed.emit(self.get_value())


class DictEdit(ValuedContainerWidget[dict]):
    def __init__(self, options: dict[str, dict], value=Undefined, **kwargs):
        from himena.qt.magicgui import get_type_map

        type_map = get_type_map()
        self._widget_dict: dict[str, ValueWidget] = {}
        for k, opt in options.items():
            _opt = opt.copy()
            if (label := _opt.pop("label", None)) is None:
                label = k
            value = _opt.pop("value", Undefined)
            annotation = _opt.pop("annotation", None)
            self._widget_dict[k] = type_map.create_widget(
                value, annotation=annotation, options=_opt, label=label
            )
        super().__init__(value, widgets=self._widget_dict.values(), **kwargs)
        for widget in self._widget_dict.values():
            widget.changed.connect(self._emit_value_changed)

    def get_value(self) -> dict:
        return {k: v.value for k, v in self._widget_dict.items()}

    def set_value(self, value: dict):
        if value is None or value is Undefined:
            value = {}
        with self.changed.blocked():
            for k, v in value.items():
                if k in self._widget_dict:
                    self._widget_dict[k].value = v
        self.changed.emit(self.get_value())

    def _emit_value_changed(self) -> None:
        self.changed.emit(self.get_value())



from __future__ import annotations

from concurrent.futures import Future
import timeit
from typing import Callable, Any, TypeVar, get_origin
import inspect
from functools import wraps
import warnings
from textwrap import dedent
from cmap import Colormap, Color

from himena.consts import StandardType
from himena.types import (
    FutureInfo,
    ParametricWidgetProtocol,
    WidgetDataModel,
    Parametric,
    ModelTrack,
    GuiConfiguration,
)
from himena.utils.misc import lru_cache
from himena.workflow import (
    CommandExecution,
    ModelParameter,
    parse_parameter,
    WindowParameter,
)
from himena.workflow import Workflow

_F = TypeVar("_F", bound=Callable)


def _get_type_arg(func: Callable, target: type) -> type | None:
    annots = [v for k, v in func.__annotations__.items() if k != "return"]
    if len(annots) != 1:
        return None
    annot = annots[0]
    if not (hasattr(annot, "__origin__") and hasattr(annot, "__args__")):
        return None
    if annot.__origin__ is not target:
        return None
    if len(annot.__args__) != 1:
        return None
    return annot.__args__[0]


def get_widget_data_model_type_arg(func: Callable) -> type | None:
    return _get_type_arg(func, WidgetDataModel)


def has_widget_data_model_argument(func: Callable) -> bool:
    """If true, the function has a WidgetDataModel type hint."""
    for k, v in func.__annotations__.items():
        if k == "return":
            continue
        if v is WidgetDataModel:
            return True
        if hasattr(v, "__origin__") and hasattr(v, "__args__"):
            if v.__origin__ is WidgetDataModel:
                return True
    return False


def get_subwindow_type_arg(func: Callable) -> type | None:
    from himena.widgets import SubWindow

    return _get_type_arg(func, SubWindow)


def get_user_context(widget: Any) -> Any:
    """Get the user context from the widget."""
    if user_context := getattr(widget, "user_context", None):
        return user_context()
    return None


@lru_cache
def get_widget_class_id(cls: type) -> str:
    """Get the widget ID from the class.

    Widget ID is always determined by the register_widget_class decorator. This ID is
    used during the application to identify the widget class.
    """
    import importlib

    if _widget_id := getattr(cls, "__himena_widget_id__", None):
        if not isinstance(_widget_id, str):
            raise TypeError(f"Widget ID must be a string, got {type(_widget_id)}")
        return _widget_id

    name = f"{cls.__module__}.{cls.__name__}"
    # look for simpler import path
    submods = cls.__module__.split(".")
    for i in range(1, len(submods)):
        mod_name = ".".join(submods[:i])
        try:
            mod = importlib.import_module(mod_name)
            if getattr(mod, cls.__name__, None) is cls:
                name = f"{mod_name}.{cls.__name__}"
                break
        except Exception:
            pass

    # replace the first "." with ":" to make names consistent
    name = name.replace(".", ":", 1)
    return name


def get_display_name(cls: type, sep: str = "\n", class_id: bool = True) -> str:
    if title := getattr(cls, "__himena_display_name__", None):
        if not isinstance(title, str):
            raise TypeError(f"Display name must be a string, got {type(title)}")
    else:
        title = cls.__name__
    name = get_widget_class_id(cls)
    if class_id:
        return f"{title}{sep}({name})"
    else:
        return title


def _is_widget_data_model(a):
    return WidgetDataModel in (get_origin(a), a)


def _is_subwindow(a):
    from himena.widgets._wrapper import SubWindow

    return SubWindow in (get_origin(a), a)


def _is_parametric(a):
    return a in (
        Parametric,
        "Parametric",
        ParametricWidgetProtocol,
        "ParametricWidgetProtocol",
    )


def make_function_callback(
    f: _F,
    command_id: str,
    title: str | None = None,
    run_async: bool = False,
) -> _F:
    from himena.widgets import SubWindow, current_instance

    try:
        sig = inspect.signature(f)
    except Exception:
        warnings.warn(f"Failed to get signature of {f!r}")
        return f

    f_annot = f.__annotations__
    keys_model: list[str] = []
    keys_subwindow: list[str] = []
    for key, param in sig.parameters.items():
        if _is_widget_data_model(param.annotation):
            keys_model.append(key)
        elif _is_subwindow(param.annotation):
            keys_subwindow.append(key)

    for key in keys_model:
        f_annot[key] = WidgetDataModel

    if _is_widget_data_model(sig.return_annotation):
        f_annot["return"] = WidgetDataModel
    elif _is_subwindow(sig.return_annotation):
        f_annot["return"] = SubWindow
    elif _is_parametric(sig.return_annotation):
        f_annot["return"] = Parametric
    elif sig.return_annotation is ParametricWidgetProtocol:
        f_annot["return"] = ParametricWidgetProtocol
    else:
        return f

    is_parametric = _is_parametric(f_annot.get("return"))

    @wraps(f)
    def _new_f(*args, **kwargs):
        bound = sig.bind(*args, **kwargs)
        _time_before = timeit.default_timer()
        out = f(*args, **kwargs)
        contexts: list[ModelParameter | WindowParameter] = []
        workflows = []
        for key, input_ in bound.arguments.items():
            input_param, wf = parse_parameter(key, input_)
            if isinstance(input_param, (ModelParameter, WindowParameter)):
                contexts.append(input_param)
                workflows.append(wf)
        workflow = Workflow.concat(workflows)
        if isinstance(out, WidgetDataModel):
            out.workflow = workflow.with_step(
                CommandExecution(
                    contexts=contexts,
                    command_id=command_id,
                    parameters=None,
                    execution_time=timeit.default_timer() - _time_before,
                )
            )
            if out.update_inplace and contexts:
                ui = current_instance()
                input_window = ui._window_for_workflow_id(contexts[0].value)
                input_window.update_model(out)
                input_window._update_model_workflow(out.workflow)
                return None
        elif is_parametric:
            ModelTrack(
                contexts=contexts,
                command_id=command_id,
                workflow=workflow,
                time_start=_time_before,
            ).set(out)
            if title is not None:
                cfg = GuiConfiguration.get(out) or GuiConfiguration()
                cfg.title = title
                cfg.run_async = run_async
                cfg.set(out)
        return out

    if run_async and not is_parametric:
        _new_f = _wrap_future(_new_f)
    return _new_f


def _wrap_future(f):
    from himena.widgets import current_instance

    return_annot = f.__annotations__.get("return")

    @wraps(f)
    def _f(*args, **kwargs):
        ins = current_instance()
        future = ins._executor.submit(f, *args, **kwargs)
        FutureInfo(return_annot).set(future)
        return future

    _f.__annotations__["return"] = Future
    return _f


def get_gui_config(fn) -> dict[str, Any]:
    if isinstance(config := GuiConfiguration.get(fn), GuiConfiguration):
        out = config.asdict()
    else:
        out = {}
    if out.get("title") is None:
        if hasattr(fn, "__name__"):
            out["title"] = fn.__name__
        else:
            out["title"] = str(fn)
    return out


def import_object(full_name: str) -> Any:
    """Import object by a period-separated full name or the widget ID."""
    import importlib
    from himena.plugins import get_widget_class

    if obj := get_widget_class(full_name):
        return obj
    mod_name, func_name = full_name.replace(":", ".", 1).rsplit(".", 1)
    mod = importlib.import_module(mod_name)
    obj = getattr(mod, func_name)
    return obj


def unwrap_lazy_model(model: WidgetDataModel) -> WidgetDataModel:
    """Unwrap the lazy object if possible."""

    if model.type != StandardType.LAZY:
        raise ValueError(f"Expected a lazy object, got type {model.type}")
    if not callable(val := model.value):
        raise ValueError(
            f"Expected a callable as the value of the WidgetDataModel, got {type(val)}"
        )
    out = val()
    if not isinstance(out, WidgetDataModel):
        raise ValueError(
            f"Expected a WidgetDataModel as the return value, got {type(out)}"
        )
    return out


def to_color_or_colormap(value) -> Color | Colormap:
    if isinstance(value, (Color, Colormap)):
        return value
    if isinstance(value, str) and value.startswith("#"):
        value = Color(value)
    elif isinstance(value, dict):
        value = Colormap(value)
    else:
        try:
            value = Colormap(value)
        except Exception:
            value = Color(value)
    return value


def doc_to_whats_this(doc: str):
    lines = doc.splitlines()
    first_line = lines[0]
    others = "\n".join(lines[1:])
    doc_dedent = dedent(first_line) + "\n" + dedent(others)
    return doc_dedent



from pathlib import Path
from qtpy.QtCore import Qt
from unittest.mock import MagicMock
from himena.testing import choose_one_dialog_response
from himena_builtins.qt.explorer._widget import QExplorerWidget
from himena_builtins.qt.explorer._widget_ssh import QSSHRemoteExplorerWidget, _make_ls_args, _make_get_type_args

from pytestqt.qtbot import QtBot


def test_workspace_widget(qtbot: QtBot, himena_ui, tmpdir):
    tmpdir = Path(tmpdir)
    mock = MagicMock()
    widget = QExplorerWidget(himena_ui)
    qtbot.add_widget(widget)
    widget.open_file_requested.connect(mock)

    widget._root.set_root_path(tmpdir)  # now it's safe to move files
    widget._file_tree._make_context_menu(widget._file_tree.model().index(0, 0))
    widget._file_tree._make_drag()
    mock.assert_not_called()

    with choose_one_dialog_response(himena_ui, "Replace"):
        widget._file_tree._paste_file([Path(__file__)], tmpdir, is_copy=True)
        assert (tmpdir / Path(__file__).name).exists()
        widget._file_tree._paste_file([tmpdir / Path(__file__).name], tmpdir, is_copy=True)


    # TODO: not working ...
    # qtree = widget._workspace_tree
    # file_index = qtree.indexBelow(qtree.model().index(0, 0))
    # qtbot.mouseDClick(
    #     qtree.viewport(),
    #     QtCore.Qt.MouseButton.LeftButton,
    #     QtCore.Qt.KeyboardModifier.NoModifier,
    #     qtree.visualRect(file_index).center(),
    # )
    # mock.assert_called_once()
    # assert isinstance(mock.call_args[0][0], Path)

def test_ssh_args():
    assert _make_ls_args("HOST", "PATH") == [
        "ssh", "-p", "22", "HOST", "ls", "PATH/", "-AF"
    ]
    assert _make_get_type_args("HOST", "PATH") == [
        "ssh", "-p", "22", "HOST", "stat", "PATH", "--format='%F'"
    ]

def test_ssh_widget(qtbot: QtBot, himena_ui):
    widget = QSSHRemoteExplorerWidget(himena_ui)
    qtbot.add_widget(widget)
    widget.show()
    widget._file_list_widget._make_context_menu()
    widget._apply_filter("a")
    assert widget._filter_widget.isHidden()
    qtbot.keyClick(widget, Qt.Key.Key_F, Qt.KeyboardModifier.ControlModifier)
    assert widget._filter_widget.isVisible()
    qtbot.keyClick(widget._filter_widget, Qt.Key.Key_Escape)
    assert widget._filter_widget.isHidden()
    widget._file_list_widget._save_items([])



from __future__ import annotations

from types import MappingProxyType
from typing import Callable, overload, TypeVar, TYPE_CHECKING
import warnings
from app_model.types import Action
from himena._descriptors import NoNeedToSave
from himena._utils import get_display_name, get_widget_class_id
from himena._app_model import AppContext as ctx
from himena.plugins.actions import AppActionRegistry, PluginConfigTuple
from himena.plugins._utils import type_to_expression

from himena.types import WidgetDataModel

if TYPE_CHECKING:
    from himena.widgets import SubWindow, MainWindow
    from himena.plugins.actions import PluginConfigType

_T = TypeVar("_T")
_WIDGET_ID_TO_WIDGET_CLASS: dict[str, type] = {}


def get_widget_class(id: str) -> type | None:
    return _WIDGET_ID_TO_WIDGET_CLASS.get(id)


@overload
def register_widget_class(
    type_: str,
    widget_class: _T,
    priority: int = 100,
    plugin_configs: PluginConfigType | None = None,
) -> _T: ...


@overload
def register_widget_class(
    type_: str,
    widget_class: None,
    priority: int = 100,
    plugin_configs: PluginConfigType | None = None,
) -> Callable[[_T], _T]: ...


def register_widget_class(type_, widget_class=None, priority=100, plugin_configs=None):
    """Register a frontend widget class for the given model type.

    The `__init__` method of the registered class must not take any argument. The class
    must implement `update_model` method to update the widget state from a
    WidgetDataModel.

    ``` python
    @register_widget("text")
    class MyTextEdit(QtW.QPlainTextEdit):
        def update_model(self, model: WidgetDataModel):
            self.setPlainText(model.value)
    ```

    There are other method names that can be implemented to make the widget more
    functional.

    - `to_model(self) -> WidgetDataModel`:
    - `model_type(self) -> str`:
    - `control_widget(self) -> <widget>`:
    - `is_modified(self) -> bool`:
    - `set_modified(self, modified: bool)`:
    - `size_hint(self) -> tuple[int, int]`:
    - `is_editable(self) -> bool`:
    - `set_editable(self, editable: bool)`:
    - `dropped_callback(self, other: WidgetDataModel)`:
    - `allowed_drop_types(self) -> list[str]`:
    - `display_name(cls) -> str`:
    - `theme_changed_callback(self, theme: Theme)`:
    - `widget_activated_callback(self)`:
    - `widget_closed_callback(self)`:
    - `widget_resized_callback(self, size_old, size_new)`:
    """

    def inner(wcls):
        import himena.qt

        widget_id = get_widget_class_id(wcls)
        is_multi_registration = False
        if existing_class := _WIDGET_ID_TO_WIDGET_CLASS.get(widget_id):
            if existing_class is wcls:
                is_multi_registration = True
            else:
                raise ValueError(
                    f"Widget class with ID {widget_id!r} already assigned for "
                    f"{existing_class}; you must assign a unique ID for each class."
                )
        _WIDGET_ID_TO_WIDGET_CLASS[widget_id] = wcls
        himena.qt.register_widget_class(type_, wcls, priority=priority)
        fn = OpenDataInFunction(type_, wcls)
        reg = AppActionRegistry.instance()
        reg.add_action(fn.to_action(), is_dynamic=True)
        if not is_multi_registration:
            wcls.__himena_model_type__ = type_

        if plugin_configs:
            cfg_type = type(plugin_configs)
            if widget_id in reg._plugin_default_configs:
                warnings.warn(
                    f"Plugin config for {widget_id!r} already registered; "
                    f"overwriting with new config {plugin_configs}.",
                    UserWarning,
                    stacklevel=2,
                )
            reg._plugin_default_configs[widget_id] = PluginConfigTuple(
                get_display_name(wcls, sep=" ", class_id=False),
                plugin_configs,
                cfg_type,
            )
        return wcls

    return inner if widget_class is None else inner(widget_class)


def widget_classes() -> MappingProxyType[str, type]:
    """Get the mapping of widget ID to widget class."""
    from himena.qt.registry._api import _APP_TYPE_TO_QWIDGET

    out = {}
    for widget_list in _APP_TYPE_TO_QWIDGET.values():
        for item in widget_list:
            out[item.type] = item.widget_class
    return MappingProxyType(out)


def register_previewer_class(type_: str, widget_class: type):
    """Register a widget class for previewing the given model type."""

    def inner(wcls):
        import himena.qt

        widget_id = get_widget_class_id(wcls)
        if existing_class := _WIDGET_ID_TO_WIDGET_CLASS.get(widget_id):
            raise ValueError(
                f"Widget class with ID {widget_id!r} already exists ({existing_class})."
            )
        _WIDGET_ID_TO_WIDGET_CLASS[widget_id] = wcls
        himena.qt.register_widget_class(type_, wcls, priority=-10)
        fn = OpenDataInFunction(type_, wcls)
        AppActionRegistry.instance().add_action(fn.to_action(), is_dynamic=True)
        fn = PreviewDataInFunction(type_, wcls)
        AppActionRegistry.instance().add_action(fn.to_action(), is_dynamic=True)
        return type_

    return inner if widget_class is None else inner(widget_class)


class OpenDataInFunction:
    """Callable class for 'open this data in ...' action."""

    def __init__(self, type_: str, widget_class: type):
        self._display_name = get_display_name(widget_class)
        self._plugin_id = get_widget_class_id(widget_class)
        self._type = type_
        self._enablement = (
            # disable action if the model data type is different
            type_to_expression(self._type)
            # disable action that open the same data in the same widget
            & (ctx.active_window_widget_id != self._plugin_id)
        )

    def __call__(self, model: WidgetDataModel) -> WidgetDataModel:
        return model.with_open_plugin(
            self._plugin_id, save_behavior_override=NoNeedToSave()
        )

    def to_action(self) -> Action:
        return Action(
            id=f"open-in:{self._plugin_id}:{self._type}",
            title=self._display_name,
            tooltip=f"Open this data in {self._display_name}",
            callback=self,
            enablement=self._enablement,
            menus=[{"id": f"/open-in/{self._type}", "group": "open-in"}],
        )


class PreviewDataInFunction:
    """Callable class for 'preview this data in ...' action."""

    def __init__(self, type_: str, widget_class: type):
        self._display_name = get_display_name(widget_class)
        self._plugin_id = get_widget_class_id(widget_class)
        self._type = type_

    def __call__(self, win: SubWindow, ui: MainWindow):
        model = win.to_model().with_open_plugin(self._plugin_id)
        previewer = ui.add_data_model(model)
        previewer._switch_to_file_watch_mode()
        return None

    def menu_id(self) -> str:
        return f"/model_menu:{self._type}/preview-in"

    def to_action(self) -> Action:
        tooltip = f"Preview this data in {self._display_name}"
        return Action(
            id=f"preview-in:{self._plugin_id}:{self._type}",
            title=self._display_name,
            tooltip=tooltip,
            callback=self,
            menus=[{"id": self.menu_id(), "group": "open-in"}],
        )



from himena import StandardType
from himena.plugins import register_widget_class, register_previewer_class
from himena_builtins.qt.widgets.text import QTextEdit, QRichTextEdit, TextEditConfigs
from himena_builtins.qt.widgets.text_previews import QSvgPreview, QMarkdownPreview

register_widget_class(StandardType.TEXT, QTextEdit, plugin_configs=TextEditConfigs())
register_widget_class(StandardType.HTML, QRichTextEdit)
register_previewer_class(StandardType.SVG, QSvgPreview)
register_previewer_class(StandardType.MARKDOWN, QMarkdownPreview)



from concurrent.futures import Future
import datetime
from pathlib import Path
from logging import getLogger
from app_model.types import (
    KeyBindingRule,
    KeyCode,
    KeyMod,
    KeyChord,
    StandardKeyBinding,
)
from himena.data_wrappers import wrap_array
from himena._descriptors import SaveToNewPath
from himena.consts import StandardType, MenuId
from himena.plugins import configure_gui, ReaderPlugin
from himena.standards.model_meta import ImageMeta
from himena.widgets import MainWindow, SubWindow
from himena import _providers, workflow as _wf
from himena.types import (
    ClipboardDataModel,
    Parametric,
    WidgetDataModel,
)
from himena._app_model._context import AppContext as _ctx
from himena._app_model.actions._registry import ACTIONS, SUBMENUS
from himena.utils.collections import OrderedSet
from himena.exceptions import Cancelled

_CtrlK = KeyMod.CtrlCmd | KeyCode.KeyK
_LOGGER = getLogger(__name__)

READ_GROUP = "0000_io_read"
WRITE_GROUP = "0001_io_write"
SCR_SHOT_GROUP = "2001_screenshot"
SETTINGS_GROUP = "3001_settings"
COPY_SCR_SHOT = "0000_copy-screenshot"
SAVE_SCR_SHOT = "0001_save-screenshot"
EXIT_GROUP = "9900_exit"


@ACTIONS.append_from_fn(
    id="open-file",
    title="Open File(s) ...",
    icon="material-symbols:folder-open-outline",
    menus=[
        {"id": MenuId.FILE, "group": READ_GROUP},
        {"id": MenuId.TOOLBAR, "group": READ_GROUP},
        {"id": MenuId.STARTUP, "group": READ_GROUP},
    ],
    keybindings=[StandardKeyBinding.Open],
)
def open_file_from_dialog(ui: MainWindow) -> Future:
    """Open file(s). Multiple files will be opened as separate sub-windows."""
    if result := ui.exec_file_dialog(mode="rm"):
        return ui.read_files_async(result)
    raise Cancelled


def _get_reader_options(file_path: Path) -> dict:
    _store = _providers.ReaderStore.instance()
    readers = _store.get(file_path, min_priority=-float("inf"))

    # prepare reader plugin choices
    choices_reader = sorted(
        [(f"{r.__name__}\n({r.plugin_str})", r) for r in readers],
        key=lambda x: x[1].priority,
        reverse=True,
    )
    return {
        "choices": choices_reader,
        "widget_type": "RadioButtons",
        "value": choices_reader[0][1],
    }


def _open_file_using_reader(
    file_path,
    reader: ReaderPlugin,
    ui: MainWindow | None = None,
    editable: bool = True,
) -> WidgetDataModel:
    model = reader.read_and_update_source(file_path)
    plugin_str = reader.plugin_str
    if ui:
        ui._recent_manager.append_recent_files([(file_path, plugin_str)])
    wf = _wf.LocalReaderMethod(
        path=file_path,
        plugin=plugin_str,
        output_model_type=model.type,
    ).construct_workflow()
    model.workflow = wf
    model.editable = editable
    return model


@ACTIONS.append_from_fn(
    id="open-file-with",
    title="Open File With ...",
    menus=[
        {"id": MenuId.FILE, "group": READ_GROUP},
        {"id": MenuId.STARTUP, "group": READ_GROUP},
    ],
    keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.KeyO)],
    run_async=True,
)
def open_file_using_from_dialog(ui: MainWindow) -> Parametric:
    """Open file using selected plugin."""
    from himena.plugins import configure_gui

    if (file_path := ui.exec_file_dialog(mode="r")) is None:
        raise Cancelled

    @configure_gui(reader=_get_reader_options(file_path))
    def choose_a_plugin(reader: ReaderPlugin) -> WidgetDataModel:
        _LOGGER.info("Reading file %s using %r", file_path, reader)
        return _open_file_using_reader(file_path, reader, ui)

    return choose_a_plugin


@ACTIONS.append_from_fn(
    id="open-file-group",
    title="Open File Group ...",
    menus=[{"id": MenuId.FILE, "group": READ_GROUP}],
)
def open_file_group_from_dialog(ui: MainWindow) -> Future:
    """Open file group as a single sub-window."""
    if result := ui.exec_file_dialog(mode="rm"):
        return ui.read_files_async(result)
    raise Cancelled


@ACTIONS.append_from_fn(
    id="open-folder",
    title="Open Folder ...",
    icon="material-symbols:folder-open",
    menus=[
        {"id": MenuId.FILE, "group": READ_GROUP},
        {"id": MenuId.STARTUP, "group": READ_GROUP},
    ],
    keybindings=[
        KeyBindingRule(primary=KeyChord(_CtrlK, KeyMod.CtrlCmd | KeyCode.KeyO))
    ],
)
def open_folder_from_dialog(ui: MainWindow) -> Future:
    """Open a folder as a sub-window."""
    if path := ui.exec_file_dialog(mode="d"):
        return ui.read_files_async([path])
    raise Cancelled


@ACTIONS.append_from_fn(
    id="watch-file-using",
    title="Watch File ...",
    menus=[{"id": MenuId.FILE, "group": READ_GROUP}],
    need_function_callback=True,
)
def watch_file_using_from_dialog(ui: MainWindow) -> Parametric:
    """Watch file using selected plugin."""
    from himena.plugins import configure_gui

    if (file_path := ui.exec_file_dialog(mode="r")) is None:
        raise Cancelled

    @configure_gui(reader=_get_reader_options(file_path))
    def choose_a_plugin(reader: ReaderPlugin) -> None:
        _LOGGER.info("Watch file %s using %r", file_path, reader)
        model = _open_file_using_reader(file_path, reader, editable=False)
        win = ui.add_data_model(model)
        win._switch_to_file_watch_mode()
        return None

    return choose_a_plugin


@ACTIONS.append_from_fn(
    id="save",
    title="Save ...",
    icon="material-symbols:save-outline",
    menus=[
        {"id": MenuId.FILE, "group": WRITE_GROUP},
        {"id": MenuId.TOOLBAR, "group": WRITE_GROUP},
    ],
    keybindings=[StandardKeyBinding.Save],
    enablement=_ctx.is_active_window_supports_to_model,
)
def save_from_dialog(ui: MainWindow, sub_win: SubWindow) -> Future:
    """Save (overwrite) the current sub-window as a file."""
    if cb := sub_win._save_from_dialog(ui):
        return ui._executor.submit(cb)
    raise Cancelled


@ACTIONS.append_from_fn(
    id="save-as",
    title="Save As ...",
    icon="material-symbols:save-as-outline",
    menus=[{"id": MenuId.FILE, "group": WRITE_GROUP}],
    keybindings=[StandardKeyBinding.SaveAs],
    enablement=_ctx.is_active_window_supports_to_model,
)
def save_as_from_dialog(ui: MainWindow, sub_win: SubWindow) -> Future:
    """Save the current sub-window as a new file."""
    if cb := sub_win._save_from_dialog(ui, behavior=SaveToNewPath()):
        return ui._executor.submit(cb)
    raise Cancelled


@ACTIONS.append_from_fn(
    id="save-as-using",
    title="Save As Using ...",
    menus=[{"id": MenuId.FILE, "group": WRITE_GROUP}],
    need_function_callback=True,
    enablement=_ctx.is_active_window_supports_to_model,
)
def save_as_using_from_dialog(ui: MainWindow, sub_win: SubWindow) -> Future:
    """Save the current sub-window using selected plugin."""
    model = sub_win.to_model()
    ins = _providers.WriterStore().instance()
    model.title = sub_win.title
    save_path = sub_win._save_behavior._determine_save_path(model) or "~"
    writers = ins.get(model, Path(save_path), min_priority=-float("inf"))

    # prepare reader plugin choices
    choices_writer = [(f"{w.__name__}\n({w.plugin.name})", w) for w in writers]

    writer = ui.exec_choose_one_dialog(
        title="Choose a plugin",
        message="Choose a plugin to save the file.",
        choices=choices_writer,
        how="radiobuttons",
    )
    if writer is None:
        raise Cancelled  # no choice selected
    if cb := sub_win._save_from_dialog(
        ui, behavior=SaveToNewPath(), plugin=writer.plugin
    ):
        return ui._executor.submit(cb)
    else:
        raise Cancelled


@ACTIONS.append_from_fn(
    id="open-recent",
    title="Open Recent ...",
    menus=[
        {"id": MenuId.FILE_RECENT, "group": READ_GROUP},
        {"id": MenuId.TOOLBAR, "group": READ_GROUP},
    ],
    keybindings=[
        KeyBindingRule(primary=KeyChord(_CtrlK, KeyMod.CtrlCmd | KeyCode.KeyR))
    ],
    icon="mdi:recent",
)
def open_recent(ui: MainWindow) -> WidgetDataModel:
    """Open a recent file as a sub-window."""
    return ui._backend_main_window._show_command_palette("recent")


@ACTIONS.append_from_fn(
    id="new",
    title="New Data...",
    menus=[
        {"id": MenuId.FILE_NEW, "group": "02_more", "order": 99},
        {"id": MenuId.STARTUP, "group": READ_GROUP},
    ],
    keybindings=[StandardKeyBinding.New],
    recording=False,
)
def open_new(ui: MainWindow) -> WidgetDataModel:
    """Open a new data as a sub-window."""
    return ui._backend_main_window._show_command_palette("new")


@ACTIONS.append_from_fn(
    id="paste-as-window",
    title="Paste As Window",
    menus=[MenuId.FILE_NEW],
    enablement=~_ctx.is_subwindow_focused,
    keybindings=[StandardKeyBinding.Paste],
)
def paste_from_clipboard(ui: MainWindow) -> WidgetDataModel:
    """Paste the clipboard data as a sub-window."""
    if data := ui._backend_main_window._clipboard_data():
        title = "Clipboard"
        if (image := data.image) is not None:
            shape = wrap_array(image).shape
            if len(shape) == 3 and shape[-1] in (3, 4):
                meta = ImageMeta(axes=["y", "x", "c"], is_rgb=True)
            else:
                meta = None
            return WidgetDataModel(
                value=image, type=StandardType.IMAGE, title=title, metadata=meta
            )
        elif files := data.files:
            ui.read_files(files)
            return None
        elif html := data.html:
            return WidgetDataModel(value=html, type=StandardType.HTML, title=title)
        elif text := data.text:
            return WidgetDataModel(value=text, type=StandardType.TEXT, title=title)
        raise ValueError("No data to paste from clipboard.")
    raise Cancelled


### Load/save session


@ACTIONS.append_from_fn(
    id="load-session",
    title="Load Session ...",
    menus=[
        {"id": MenuId.FILE_SESSION, "group": READ_GROUP},
        {"id": MenuId.STARTUP, "group": READ_GROUP},
    ],
    keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.KeyL)],
)
def load_session_from_dialog(ui: MainWindow) -> None:
    """Load a application session from a file."""
    if path := ui.exec_file_dialog(
        mode="r",
        allowed_extensions=[".session.zip"],
        group="session",
    ):
        ui.load_session(path)
    raise Cancelled


@ACTIONS.append_from_fn(
    id="save-session",
    title="Save Session ...",
    menus=[{"id": MenuId.FILE_SESSION, "group": WRITE_GROUP}],
    enablement=_ctx.num_tabs > 0,
    run_async=True,
)
def save_session(ui: MainWindow) -> Parametric:
    datetime_str = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    command_ids = OrderedSet[str]()
    choices: list[tuple[str, str]] = []
    for win in ui.iter_windows():
        if isinstance(step := win._widget_workflow.last(), _wf.CommandExecution):
            if step.command_id in command_ids:
                continue
            if cmd := ui.model_app.registered_actions.get(step.command_id):
                command_ids.add(cmd.id)
                choices.append((f"{cmd.title} ({cmd.id})", cmd.id))

    @configure_gui(
        save_path={
            "mode": "w",
            "filter": "Session file (*.session.zip);;Session directory (*.session;*)",
            "value": f"himena-{datetime_str}.session.zip",
        },
        allow_calculate={"choices": choices, "widget_type": "Select"},
    )
    def run_save_session(
        save_path: Path,
        save_copies: bool = False,
        allow_calculate: list[str] = (),
    ) -> None:
        return ui.save_session(
            save_path,
            save_copies=save_copies,
            allow_calculate=allow_calculate,
        )

    return run_save_session


@ACTIONS.append_from_fn(
    id="save-tab-session",
    title="Save Tab Session ...",
    menus=[{"id": MenuId.FILE_SESSION, "group": WRITE_GROUP}],
    enablement=(_ctx.num_tabs > 0) & (_ctx.num_sub_windows > 0),
)
def save_tab_session_from_dialog(ui: MainWindow) -> None:
    """Save current application state to a session."""
    datetime_str = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    if path := ui.exec_file_dialog(
        mode="w",
        extension_default=".session.yaml",
        allowed_extensions=[".session.yaml"],
        start_path=f"Tab-{datetime_str}.session.yaml",
        group="session",
    ):
        if tab := ui.tabs.current():
            return tab.save_session(path)
    raise Cancelled


@ACTIONS.append_from_fn(
    id="builtins:exec-workflow-file",
    title="Execute Workflow File ...",
    menus=[{"id": MenuId.FILE, "group": READ_GROUP}],
    need_function_callback=True,
)
def exec_workflow_file(ui: MainWindow) -> Parametric:
    """Execute a workflow from a workflow file."""
    if path := ui.exec_file_dialog(
        extension_default=".workflow.json",
        allowed_extensions=[".txt", ".json", ".workflow.json"],
        caption="Select a workflow file",
        group="workflows",
    ):
        return _wf.as_function_from_path(path)(ui)
    raise Cancelled


@ACTIONS.append_from_fn(
    id="quit",
    title="Quit",
    menus=[{"id": MenuId.FILE, "group": EXIT_GROUP}],
    keybindings=[StandardKeyBinding.Quit],
    recording=False,
)
def quit_main_window(ui: MainWindow) -> None:
    """Quit the application."""
    ui._backend_main_window._exit_main_window(confirm=True)


@ACTIONS.append_from_fn(
    id="copy-screenshot",
    title="Copy Screenshot of Entire Main Window",
    menus=[{"id": MenuId.FILE_SCREENSHOT, "group": COPY_SCR_SHOT}],
)
def copy_screenshot(ui: MainWindow) -> ClipboardDataModel:
    """Copy a screenshot of the main window to the clipboard."""
    data = ui._backend_main_window._screenshot("main")
    return ClipboardDataModel(image=data)


@ACTIONS.append_from_fn(
    id="copy-screenshot-area",
    title="Copy Screenshot of Tab Area",
    menus=[{"id": MenuId.FILE_SCREENSHOT, "group": COPY_SCR_SHOT}],
    enablement=_ctx.num_tabs > 0,
)
def copy_screenshot_area(ui: MainWindow) -> ClipboardDataModel:
    """Copy a screenshot of the tab area to the clipboard."""
    data = ui._backend_main_window._screenshot("area")
    return ClipboardDataModel(image=data)


@ACTIONS.append_from_fn(
    id="copy-screenshot-window",
    title="Copy Screenshot of Sub-Window",
    menus=[{"id": MenuId.FILE_SCREENSHOT, "group": COPY_SCR_SHOT}],
    enablement=_ctx.num_sub_windows > 0,
)
def copy_screenshot_window(ui: MainWindow) -> ClipboardDataModel:
    """Copy a screenshot of the sub window to the clipboard."""
    data = ui._backend_main_window._screenshot("window")
    return ClipboardDataModel(image=data)


@ACTIONS.append_from_fn(
    id="settings",
    title="Settings ...",
    menus=[
        {"id": MenuId.FILE, "group": SETTINGS_GROUP},
        {"id": MenuId.CORNER, "group": SETTINGS_GROUP},
    ],
    keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.Comma)],
    icon="weui:setting-filled",
)
def show_setting_dialog(ui: MainWindow):
    """Open a dialog to edit the application profile."""
    from himena.qt.settings import QSettingsDialog

    return QSettingsDialog(ui).exec()


def _save_screenshot(ui: MainWindow, target: str) -> None:
    from PIL import Image
    import numpy as np

    arr = ui._backend_main_window._screenshot(target)
    save_path = ui.exec_file_dialog(
        mode="w",
        extension_default=".png",
        start_path="Screenshot.png",
        group="screenshot",
    )
    if save_path is None:
        raise Cancelled
    img = Image.fromarray(np.asarray(arr))
    img.save(save_path)


@ACTIONS.append_from_fn(
    id="save-screenshot",
    title="Save Screenshot of Entire Main Window",
    menus=[{"id": MenuId.FILE_SCREENSHOT, "group": SAVE_SCR_SHOT}],
)
def save_screenshot(ui: MainWindow) -> None:
    _save_screenshot(ui, "main")


@ACTIONS.append_from_fn(
    id="save-screenshot-area",
    title="Save Screenshot of Tab Area",
    menus=[{"id": MenuId.FILE_SCREENSHOT, "group": SAVE_SCR_SHOT}],
    enablement=_ctx.num_tabs > 0,
)
def save_screenshot_area(ui: MainWindow) -> None:
    _save_screenshot(ui, "area")


@ACTIONS.append_from_fn(
    id="save-screenshot-window",
    title="Save Screenshot of Sub-Window",
    menus=[{"id": MenuId.FILE_SCREENSHOT, "group": SAVE_SCR_SHOT}],
    enablement=_ctx.num_sub_windows > 0,
)
def save_screenshot_window(ui: MainWindow) -> None:
    _save_screenshot(ui, "window")


SUBMENUS.append_from(
    id=MenuId.FILE,
    submenu=MenuId.FILE_RECENT,
    title="Open Recent",
    group=READ_GROUP,
)
SUBMENUS.append_from(
    id=MenuId.FILE,
    submenu=MenuId.FILE_SESSION,
    title="Session",
    group=READ_GROUP,
)
SUBMENUS.append_from(
    id=MenuId.FILE,
    submenu=MenuId.FILE_SCREENSHOT,
    title="Screenshot",
    group=SCR_SHOT_GROUP,
)



from himena import StandardType
from himena.plugins import register_widget_class

from himena_builtins.qt.widgets.dataframe import QDictView
from himena_builtins.qt.widgets.model_stack import QModelStack
from himena_builtins.qt.widgets.reader_not_found import QReaderNotFound
from himena_builtins.qt.widgets.function import QFunctionEdit
from himena_builtins.qt.widgets.workflow import QWorkflowView

register_widget_class(StandardType.FUNCTION, QFunctionEdit, priority=50)
register_widget_class(StandardType.MODELS, QModelStack, priority=50)
register_widget_class(StandardType.WORKFLOW, QWorkflowView, priority=50)
register_widget_class(StandardType.READER_NOT_FOUND, QReaderNotFound, priority=0)
register_widget_class(StandardType.DICT, QDictView, priority=10)



from __future__ import annotations

from pathlib import Path
from io import TextIOWrapper
import socket
from dataclasses import dataclass, asdict
import yaml
from pydantic import BaseModel, Field


def lock_file_path(name: str) -> Path:
    """Get the lock file path."""
    from himena.profile import data_dir

    dir_path = data_dir() / "lock"
    if not dir_path.exists():
        dir_path.mkdir(parents=True)
    return dir_path / f"{name}.lock"


def get_unique_lock_file(name: str) -> TextIOWrapper | None:
    """Create a lock file, return None if it already exists."""
    lock_file = lock_file_path(name)
    if lock_file.exists():
        return None
    return lock_file.open("w")


@dataclass
class SocketInfo:
    host: str = "localhost"
    port: int = 49200

    def asdict(self) -> dict[str, int | str]:
        """Convert to a dictionary."""
        return asdict(self)

    @classmethod
    def from_lock(cls, name: str) -> SocketInfo:
        """Get the socket info from the lock file."""
        with lock_file_path(name).open("r") as f:
            yml = yaml.load(f, Loader=yaml.Loader)
        if yml is None:
            yml = {}
        return SocketInfo(
            host=yml.get("host", "localhost"), port=yml.get("port", 49200)
        )

    def dump(self, path):
        yaml.dump(self.asdict(), path)

    def send_to_window(self, profile: str, files: list[str] | None = None) -> bool:
        """Send data to the window."""
        data = InterProcessData(
            profile_name=profile,
            files=files or [],
        )
        try:
            data.send(self.host, self.port)
        except Exception as e:
            print(f"Socket is not available: {e}")
            return False
        else:
            print(f"Sent data to {profile} window at {self.host}:{self.port}.")
            return True


class InterProcessData(BaseModel):
    """Data to be sent over the socket."""

    profile_name: str = Field(..., description="Name of the profile")
    files: list[str] = Field(default_factory=list, description="List of files to send")

    def to_bytes(self) -> bytes:
        """Convert the data to bytes."""
        return self.model_dump_json().encode("utf-8")

    @classmethod
    def from_bytes(cls, data: bytes) -> InterProcessData:
        """Convert bytes back to the data model."""
        return cls.model_validate_json(data.decode("utf-8"))

    def send(self, host: str = "localhost", port: int = 49200) -> None:
        """Send the data to the specified host and port using a socket."""
        with socket.create_connection((host, port)) as sock:
            sock.sendall(self.to_bytes())



from __future__ import annotations

from typing import (
    Generic,
    Hashable,
    Iterable,
    Iterator,
    MutableSet,
    Sequence,
    TypeVar,
)

_T = TypeVar("_T", bound=Hashable)


class OrderedSet(MutableSet[_T]):
    """A set that maintains the order of the elements."""

    def __init__(self, iterable: Iterable[_T] = ()):
        self._dict: dict[_T, None] = dict.fromkeys(iterable)

    def __contains__(self, other) -> bool:
        return other in self._dict

    def __iter__(self) -> Iterator[_T]:
        yield from self._dict

    def __len__(self) -> int:
        return len(self._dict)

    def add(self, value: _T) -> None:
        self._dict[value] = None

    def discard(self, value: _T) -> None:
        self._dict.pop(value, None)

    def update(self, other: Iterable[_T]) -> None:
        for value in other:
            self.add(value)


class FrozenList(Sequence[_T]):
    """A immutable list."""

    def __init__(self, iterable: Iterable[_T]):
        self._list = list(iterable)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._list!r})"

    def __getitem__(self, index: int) -> _T:
        return self._list[index]

    def __len__(self) -> int:
        return len(self._list)

    def __iter__(self) -> Iterator[_T]:
        yield from self._list


class UndoRedoStack(Generic[_T]):
    """A simple undo/redo stack to store the history."""

    def __init__(self, size: int = 10):
        self._stack_undo: list[_T] = []
        self._stack_redo: list[_T] = []
        self._size = size

    def push(self, value: _T):
        """Push a new value."""
        self._stack_undo.append(value)
        self._stack_redo.clear()
        if len(self._stack_undo) > self._size:
            self._stack_undo.pop(0)

    def undo(self) -> _T | None:
        """Undo and return the value. None if empty."""
        if len(self._stack_undo) == 0:
            return None
        value = self._stack_undo.pop()
        self._stack_redo.append(value)
        return value

    def redo(self) -> _T | None:
        """Redo and return the value. None if empty."""
        if len(self._stack_redo) == 0:
            return None
        value = self._stack_redo.pop()
        self._stack_undo.append(value)
        return value

    def undoable(self) -> bool:
        """If undo is possible."""
        return len(self._stack_undo) > 0

    def redoable(self) -> bool:
        """If redo is possible."""
        return len(self._stack_redo) > 0

    def clear(self):
        """Clear the stack."""
        self._stack_undo.clear()
        self._stack_redo.clear()



from __future__ import annotations

from typing import TypeVar, Callable
from app_model.types import Action, SubmenuItem

from himena._utils import make_function_callback
from himena.consts import NO_RECORDING_FIELD

_F = TypeVar("_F", bound=Callable)


class ActionList(list[Action]):
    def append_from_fn(
        self,
        id: str,
        title: str | None = None,
        icon: str | None = None,
        menus=None,
        enablement=None,
        keybindings=None,
        need_function_callback: bool = False,
        run_async: bool = False,
        recording: bool = True,  # whether to add the the command history
    ) -> Callable[[_F], _F]:
        def inner(fn: _F) -> _F:
            if need_function_callback or run_async:
                callback = make_function_callback(
                    fn, id, title=title, run_async=run_async
                )
            else:
                callback = fn
            if not recording:
                setattr(callback, NO_RECORDING_FIELD, True)
            action = Action(
                id=id,
                title=title or id,
                icon=icon,
                callback=callback,
                tooltip=fn.__doc__,
                status_tip=fn.__doc__,
                menus=menus,
                keybindings=keybindings,
                enablement=enablement,
                icon_visible_in_menu=False,
            )
            self.append(action)
            return fn

        return inner


class SubmenuList(list[tuple[str, SubmenuItem]]):
    def append_from(
        self,
        id: str,
        submenu: str,
        title: str,
        enablement=None,
        group: str | None = None,
    ) -> SubmenuList:
        self.append(
            (
                id,
                SubmenuItem(
                    submenu=submenu, title=title, enablement=enablement, group=group
                ),
            )
        )
        return self


ACTIONS = ActionList()
SUBMENUS = SubmenuList()



import numpy as np
from qtpy.QtCore import Qt, QPoint
from pytestqt.qtbot import QtBot
from himena import MainWindow, StandardType, WidgetDataModel, _drag
from himena.testing import WidgetTester
from himena.standards import roi
from himena_builtins.qt.basic import QModelStack

def test_model_stack_widget(himena_ui: MainWindow, qtbot: QtBot):
    win = himena_ui.add_object(
        {"model-0": WidgetDataModel(value="a", type=StandardType.TEXT)},
        type=StandardType.MODELS,
    )
    win.update_model(
        [
            ("model-0", WidgetDataModel(value="a", type=StandardType.TEXT)),
            ("model-1", WidgetDataModel(value="a", type=StandardType.TEXT)),
        ],
        type=StandardType.MODELS,
    )

    assert isinstance(win.widget, QModelStack)
    with WidgetTester(win.widget) as tester:
        tester.widget.show()
        mlist = tester.widget._model_list
        old, new = tester.cycle_model()
        assert len(old.value) == len(new.value)
        assert mlist.count() == 2
        tester.drop_model(
            [
                ("model-10", WidgetDataModel(value="a", type=StandardType.TEXT)),
                ("model-11", WidgetDataModel(value="a", type=StandardType.TEXT)),
            ],
            type=StandardType.MODELS,
        )
        assert mlist.count() == 3  # model list is dropped as a nested list
        row_1 = mlist.model().index(1, 0)
        point_1 = mlist.visualRect(row_1).center()
        qtbot.mouseClick(mlist.viewport(), Qt.MouseButton.LeftButton, pos=point_1)
        assert mlist.currentRow() == 1
        qtbot.mouseMove(mlist.viewport(), point_1 + QPoint(2, 2))

def test_commands(himena_ui: MainWindow):
    win0 = himena_ui.add_object(value="abc", type=StandardType.TEXT)
    win1 = himena_ui.add_object(value=[[1, 2], [3, 2]], type=StandardType.TABLE)
    himena_ui.exec_action("builtins:models:stack-models", with_params={"models": [win1.to_model(), win0.to_model()]})
    win2 = himena_ui.current_window
    assert isinstance(win2.widget, QModelStack)
    assert win2.widget._model_list.count() == 2
    himena_ui.exec_action("builtins:models:sort-model-list", with_params={"sort_by": "title"})
    himena_ui.exec_action("builtins:models:sort-model-list", with_params={"sort_by": "type"})
    himena_ui.exec_action("builtins:models:sort-model-list", with_params={"sort_by": "time"})
    himena_ui.exec_action("builtins:models:filter-model-list", with_params={"model_type": "text"})
    himena_ui.exec_action("builtins:models:filter-model-list", with_params={"title_contains": "X"})
    himena_ui.exec_action("builtins:models:compute-lazy-items")

def test_events(himena_ui: MainWindow, qtbot: QtBot):
    stack = QModelStack(himena_ui)
    himena_ui.add_widget(stack)
    qtbot.addWidget(stack)
    with WidgetTester(stack) as tester:
        tester.update_model(
            {"model-0": WidgetDataModel(value=np.zeros((2, 2)), type=StandardType.IMAGE),
             "model-1": WidgetDataModel(value="a", type=StandardType.TEXT)}
        )
        tester.widget._model_list.setCurrentRow(0)
        rois = roi.RoiListModel(items=[roi.LineRoi(start=(0, 0), end=(1, 1))])
        _drag.drag(WidgetDataModel(value=rois, type=StandardType.ROIS))
        stack._widget_stack._drop_event()
        stack._model_list._hover_event(stack._model_list.rect().center())
        stack._model_list._hover_event(stack._model_list.visualItemRect(stack._model_list.item(0)).center())
        stack._model_list._hover_event(QPoint(1000, 1000))



from himena._app_model.actions import (
    file_actions,
    view_actions,
    window_actions,
    tools_actions,
    go_action,
    help_actions,
)  # register actions

from himena._app_model.actions._registry import ACTIONS, SUBMENUS

del file_actions, view_actions, window_actions, tools_actions, go_action, help_actions

__all__ = ["ACTIONS", "SUBMENUS"]



from __future__ import annotations
from typing import TYPE_CHECKING, Iterator
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt


if TYPE_CHECKING:
    from ._base import QTableBase


class QHeaderViewBase(QtW.QHeaderView):
    """The header view for the QTableBase."""

    _Orientation: Qt.Orientation

    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(self._Orientation, parent)
        self.setSelectionMode(QtW.QHeaderView.SelectionMode.SingleSelection)
        self.setSectionsClickable(True)
        self.sectionPressed.connect(self._on_section_pressed)  # pressed
        self.sectionClicked.connect(self._on_section_clicked)  # released
        self.sectionEntered.connect(self._on_section_entered)  # dragged

    # fmt: off
    if TYPE_CHECKING:
        def parentWidget(self) -> QTableBase: ...
    # fmt: on

    @property
    def selection_model(self):
        return self.parentWidget()._selection_model

    def _on_section_pressed(self, logicalIndex: int) -> None:
        self.selection_model.jump_to(*self._index_for_selection_model(logicalIndex))
        self.selection_model.set_shift(True)
        return None

    def _on_section_entered(self, logicalIndex: int) -> None:
        self.selection_model.move_to(*self._index_for_selection_model(logicalIndex))
        return None

    def _on_section_clicked(self, logicalIndex) -> None:
        self.selection_model.set_shift(False)

    def _iter_selections(self) -> Iterator[slice]:
        """Iterate selections"""
        raise NotImplementedError()

    def _index_for_selection_model(self, logicalIndex: int) -> tuple[int, int]:
        raise NotImplementedError()

    def visualRectAtIndex(self, index: int) -> QtCore.QRect:
        """Return the visual rect of the given index."""
        raise NotImplementedError()

    @staticmethod
    def drawBorder(painter: QtGui.QPainter, rect: QtCore.QRect):
        """Draw the opened border of a section."""
        raise NotImplementedError()

    def drawCurrent(self, painter: QtGui.QPainter, rect: QtCore.QRect):
        """Draw the current index if exists."""
        raise NotImplementedError()

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        super().paintEvent(event)
        painter = QtGui.QPainter(self.viewport())
        color = self.parentWidget()._selection_color
        pen = QtGui.QPen(color, 3)
        painter.setPen(pen)

        # paint selections
        for _slice in self._iter_selections():
            rect_start = self.visualRectAtIndex(_slice.start)
            rect_stop = self.visualRectAtIndex(_slice.stop - 1)
            rect = rect_start | rect_stop
            self.drawBorder(painter, rect)

        # paint current
        self.drawCurrent(painter)
        return None

    def mouseReleaseEvent(self, e: QtGui.QMouseEvent) -> None:
        self.selection_model.set_shift(False)
        return super().mouseReleaseEvent(e)

    def keyPressEvent(self, e):
        self._update_press_release(e.modifiers())
        return super().keyPressEvent(e)

    def keyReleaseEvent(self, a0):
        self._update_press_release(a0.modifiers())
        return super().keyReleaseEvent(a0)

    def _update_press_release(self, mod: Qt.KeyboardModifier):
        has_ctrl = mod & Qt.KeyboardModifier.ControlModifier
        has_shift = mod & Qt.KeyboardModifier.ShiftModifier
        self.selection_model.set_shift(has_shift)
        self.selection_model.set_ctrl(has_ctrl)


class QHorizontalHeaderView(QHeaderViewBase):
    _Orientation = Qt.Orientation.Horizontal

    def visualRectAtIndex(self, index: int) -> QtCore.QRect:
        x = self.sectionViewportPosition(index)
        y = self.rect().top()
        height = self.height()
        width = self.sectionSize(index)
        return QtCore.QRect(x, y, width, height)

    @staticmethod
    def drawBorder(painter: QtGui.QPainter, rect: QtCore.QRect):
        return painter.drawPolyline(
            rect.bottomLeft(),
            rect.topLeft(),
            rect.topRight(),
            rect.bottomRight(),
        )

    def drawCurrent(self, painter: QtGui.QPainter):
        row, col = self.selection_model.current_index
        if row < 0 and col >= 0:
            rect_current = self.visualRectAtIndex(col)
            rect_current.adjust(1, 1, -1, -1)
            color = self.parentWidget()._current_color
            pen = QtGui.QPen(color, 3)
            painter.setPen(pen)
            painter.drawRect(rect_current)
        return None

    def _iter_selections(self):
        yield from self.selection_model.iter_col_selections()

    def _index_for_selection_model(self, logicalIndex):
        return -1, logicalIndex


class QVerticalHeaderView(QHeaderViewBase):
    _Orientation = Qt.Orientation.Vertical

    def visualRectAtIndex(self, index: int) -> QtCore.QRect:
        x = self.rect().left()
        y = self.sectionViewportPosition(index)
        height = self.sectionSize(index)
        width = self.width()
        return QtCore.QRect(x, y, width, height)

    @staticmethod
    def drawBorder(painter: QtGui.QPainter, rect: QtCore.QRect):
        return painter.drawPolyline(
            rect.topRight(),
            rect.topLeft(),
            rect.bottomLeft(),
            rect.bottomRight(),
        )

    def drawCurrent(self, painter: QtGui.QPainter):
        row, col = self.selection_model.current_index
        if col < 0 and row >= 0:
            rect_current = self.visualRectAtIndex(row)
            rect_current.adjust(1, 1, -1, -1)
            color = self.parentWidget()._current_color
            pen = QtGui.QPen(color, 4)
            painter.setPen(pen)
            painter.drawRect(rect_current)
        return None

    def _iter_selections(self):
        yield from self.selection_model.iter_row_selections()

    def _index_for_selection_model(self, logicalIndex):
        return logicalIndex, -1



from qtpy import QtWidgets as QtW, QtCore, QtGui
from superqt import QIconifyIcon


class QDraggableArea(QtW.QWidget):
    """A draggable area implemented with dragged event."""

    pressed = QtCore.Signal()
    dragged = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        icon = QIconifyIcon("qlementine-icons:drag-16", color="#777777")
        self._icon = icon
        self._pixmap = icon.pixmap(100, 100)

        self.setCursor(QtCore.Qt.CursorShape.SizeAllCursor)
        self._pressed = False
        self.setToolTip("Drag area")
        self.setMinimumWidth(15)

    def paintEvent(self, a0):
        painter = QtGui.QPainter(self)
        painter.setRenderHint(QtGui.QPainter.RenderHint.SmoothPixmapTransform)
        target_rect = self.rect()
        painter.drawPixmap(target_rect, self._pixmap)

    def mousePressEvent(self, a0):
        if a0.button() == QtCore.Qt.MouseButton.LeftButton:
            self._pressed = True
            self.pressed.emit()

    def mouseMoveEvent(self, a0):
        if self._pressed:
            self._pressed = False
            self.dragged.emit()

    def mouseReleaseEvent(self, a0):
        self._pressed = False



__version__ = "0.0.11"
__author__ = "Hanjin Liu"

from typing import TYPE_CHECKING
from himena.core import (
    new_window,
    create_model,
    create_array_model,
    create_dataframe_model,
    create_image_model,
    create_table_model,
    create_text_model,
)
from himena.consts import StandardType
from himena.widgets import MainWindow
from himena.types import WidgetDataModel, ClipboardDataModel, Parametric
from himena._app_model import AppContext

__all__ = [
    "new_window",
    "create_model",
    "create_array_model",
    "create_dataframe_model",
    "create_image_model",
    "create_table_model",
    "create_text_model",
    "StandardType",
    "MainWindow",
    "WidgetDataModel",
    "ClipboardDataModel",
    "Parametric",
    "AppContext",
]

if TYPE_CHECKING:
    from himena.standards import plotting  # noqa: F401


def __getattr__(name: str):
    if name == "plotting":
        from himena.standards import plotting

        return plotting
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtCore, QtGui

from himena.consts import StandardType
from himena.plugins._checker import validate_protocol
from himena.types import WidgetDataModel


class QSvgPreview(QtW.QWidget):
    """The previewer for a text editor with SVG content."""

    __himena_widget_id__ = "builtins:QSvgPreview"
    __himena_display_name__ = "Built-in SVG Preview"

    def __init__(self):
        from qtpy import QtSvg

        super().__init__()
        self._svg_renderer = QtSvg.QSvgRenderer()
        self._svg_renderer.setAspectRatioMode(QtCore.Qt.AspectRatioMode.KeepAspectRatio)
        self._svg_content: str = ""
        self._is_valid = True
        self._model_type = StandardType.SVG
        self._brush = QtGui.QBrush(
            QtGui.QColor(128, 128, 128, 128), QtCore.Qt.BrushStyle.Dense4Pattern
        )
        self._brush.setTransform(QtGui.QTransform().scale(10, 10))

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        content = str(model.value)
        if _is_valid := self._svg_renderer.load(content.encode()):
            self._svg_content = content
        else:
            self._svg_renderer.load(self._svg_content.encode())
        self._is_valid = _is_valid
        self._model_type = model.type
        self.update()

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        return WidgetDataModel(
            value=self._svg_content,
            type=self.model_type(),
        )

    @validate_protocol
    def model_type(self) -> StandardType:
        return self._model_type

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 280, 280

    def paintEvent(self, event):
        painter = QtGui.QPainter(self)
        painter.setBrush(self._brush)
        painter.drawRect(self.rect())
        self._svg_renderer.render(painter)
        if not self._is_valid:
            painter.setPen(QtGui.QPen(QtCore.Qt.GlobalColor.red, 1))
            painter.setFont(QtGui.QFont("Arial", 12))
            painter.drawText(
                self.rect().bottomLeft() + QtCore.QPoint(2, -2), "Invalid SVG"
            )
        painter.end()


class QMarkdownPreview(QtW.QWidget):
    """The previewer for a text editor with markdown content."""

    __himena_widget_id__ = "builtins:QMarkdownPreview"
    __himena_display_name__ = "Built-in Markdown Preview"

    def __init__(self):
        super().__init__()
        self._text_edit = QtW.QTextBrowser(self)
        self._text_edit.setOpenExternalLinks(True)
        self._text_edit.setOpenLinks(True)
        layout = QtW.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._text_edit)
        self._text_edit.setReadOnly(True)
        self._model_type = StandardType.MARKDOWN
        self._text_edit.setFont(QtGui.QFont("Arial", 10))

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        self._model_type = model.type
        self._text_edit.setMarkdown(model.value)

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        return WidgetDataModel(
            value=self._text_edit.toPlainText(),
            type=self.model_type(),
        )

    @validate_protocol
    def model_type(self) -> StandardType:
        return self._model_type

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return (320, 400)



from __future__ import annotations

import inspect
import weakref
from himena.types import WidgetDataModel, WindowState, WindowRect


class MockWidget:
    def __init__(self) -> None:
        self._subwindow_ref = lambda: None
        self._dockwidget_ref = lambda: None

    @property
    def subwindow(self) -> MockSubWindow:
        if win := self._subwindow_ref():
            return win
        raise ValueError("Subwindow not set")

    @property
    def dockwidget(self) -> MockDockWidget:
        if dock := self._dockwidget_ref():
            return dock
        raise ValueError("Dockwidget not set")

    def set_subwindow(self, subwindow: MockSubWindow) -> None:
        """Set the subwindow reference."""
        self._subwindow_ref = weakref.ref(subwindow)

    def set_dockwidget(self, dockwidget: MockDockWidget) -> None:
        """Set the dockwidget reference."""
        self._dockwidget_ref = weakref.ref(dockwidget)


class MockSubWindow:
    def __init__(self, widget: MockWidget, parent: MockTab, title: str = ""):
        self._widget = widget
        self._parent = weakref.ref(parent)
        self._state = WindowState.NORMAL
        self._rect = WindowRect(0, 0, 800, 600)
        self._title = title

    @property
    def tab(self) -> MockTab:
        return self._parent()


class MockTab:
    def __init__(self, title: str = ""):
        self.sub_windows: list[MockSubWindow] = []
        self.current_index: int | None = None
        self.title = title

    def __getitem__(self, index: int) -> MockSubWindow:
        return self.sub_windows[index]


class MockDockWidget:
    def __init__(self, widget: MockWidget, title: str = ""):
        self._widget = widget
        self.title = title
        self._visible = True


class MockParametricWidget(MockWidget):
    def __init__(self, sig: inspect.Signature):
        super().__init__()
        self._sig = sig

    def get_params(self):
        out = {}
        for param in self._sig.parameters.values():
            if param.default == inspect.Parameter.empty:
                raise ValueError(f"Parameter {param.name} has no default value")
            out[param.name] = param.default
        return out

    def connect_changed_signal(self, callback):
        """Do nothing."""

    def is_preview_enabled(self):
        return False


class MockModelWrapper(MockWidget):
    def __init__(self):
        super().__init__()
        self._model: WidgetDataModel | None = None
        self._editable = True

    def update_model(self, model: WidgetDataModel) -> None:
        self._model = model

    def to_model(self) -> WidgetDataModel:
        if self._model is None:
            raise ValueError("Model not set")
        return self._model.model_copy()

    def is_editable(self) -> bool:
        return self._editable

    def set_editable(self, val):
        self._editable = val



from __future__ import annotations

from pathlib import Path
from typing import Iterator
from dataclasses import dataclass
from importlib.metadata import distributions


@dataclass(frozen=True)
class HimenaPluginInfo:
    name: str
    """Plugin name, such as 'My test plugin'."""
    place: str
    """Plugin place, such as 'himena_test_plugin.io'."""
    version: str
    """Plugin version, such as '0.1.0'."""
    distribution: str
    """Distribution name, such as 'himena-test-plugin'."""

    @classmethod
    def from_local_file(cls, path: str) -> HimenaPluginInfo:
        _path = Path(path).resolve()
        return HimenaPluginInfo(
            name=_path.name,
            place=_path.as_posix(),
            version="0.0.0",
            distribution="<local>",
        )

    def is_local_plugin(self) -> bool:
        return self.distribution == "<local>"


ENTRY_POINT_GROUP_NAME = "himena.plugin"


def iter_plugin_info() -> Iterator[HimenaPluginInfo]:
    """Iterate over all the available plugins in the env."""
    for dist in distributions():
        for ep in dist.entry_points:
            if ep.group == ENTRY_POINT_GROUP_NAME:
                yield HimenaPluginInfo(
                    name=ep.name,
                    place=ep.value,
                    version=dist.version,
                    distribution=dist.name,
                )


def is_submodule(string: str, supertype: str) -> bool:
    """Check if a module is a submodule of another module.

    ``` python
    is_submodule("himena_builtins.io", "himena_builtins")  # True
    is_submodule("himena_builtins.io", "himena_builtins.io")  # True
    is_submodule("himena_builtins.io", "himena_test_plugin.io")  # False
    ```
    """
    string_parts = string.split(".")
    supertype_parts = supertype.split(".")
    if len(supertype_parts) > len(string_parts):
        return False
    return string_parts[: len(supertype_parts)] == supertype_parts



from __future__ import annotations

from pathlib import Path


def local_to_remote(
    protocol: str,
    src: Path,
    dst: str,
    is_wsl: bool = False,
    is_dir: bool = False,
    port: int = 22,
) -> list[str]:
    """Send local file to the remote host."""
    if is_wsl:
        src_wsl = to_wsl_path(src)
        args = ["wsl", "-e"] + to_command_args(
            protocol, src_wsl, dst, is_dir, port=port
        )
    else:
        args = to_command_args(protocol, src.as_posix(), dst, is_dir, port=port)
    return args


def remote_to_local(
    protocol: str,
    src: str,
    dst_path: Path,
    is_wsl: bool = False,
    is_dir: bool = False,
    port: int = 22,
) -> list[str]:
    """Run scp/rsync command to move the file from remote to local `dst_path`."""
    if is_wsl:
        dst_wsl = to_wsl_path(dst_path)
        args = ["wsl", "-e"] + to_command_args(
            protocol, src, dst_wsl, is_dir=is_dir, port=port
        )
    else:
        dst = dst_path.as_posix()
        args = to_command_args(protocol, src, dst, is_dir=is_dir, port=port)
    return args


def to_command_args(
    protocol: str,
    src: str,
    dst: str,
    is_dir: bool = False,
    port: int = 22,
) -> list[str]:
    if protocol == "rsync":
        # FIXME: f"--rsh=\"ssh -p {port}\"" should be added here, but it doesn't work
        # because of "No such file or directory (2)"
        if is_dir:
            return ["rsync", "-ar", "--progress", src, dst]
        else:
            return ["rsync", "-a", "--progress", src, dst]
    elif protocol == "scp":
        if is_dir:
            return ["scp", "-P", str(port), "-r", src, dst]
        else:
            return ["scp", "-P", str(port), src, dst]
    raise ValueError(f"Unsupported protocol {protocol!r} (must be 'rsync' or 'scp')")


def to_wsl_path(src: Path) -> str:
    """Convert an absolute Windows path to a WSL path.

    Examples
    --------
    to_wsl_path(Path("C:/Users/me/Documents")) -> "/mnt/c/Users/me/Documents"
    to_wsl_path(Path("D:/path/to/file.txt")) -> "/mnt/d/path/to/file.txt"
    """
    drive = src.drive
    drive_rel = drive + "/"
    wsl_root = Path("mnt") / drive.lower().rstrip(":")
    src_pathobj_wsl = wsl_root / src.relative_to(drive_rel).as_posix()
    return "/" + src_pathobj_wsl.as_posix()



# This file is mostly copied from pyapp-kit/ndv
# BSD-3-Clause License
# https://github.com/pyapp-kit/ndv

from __future__ import annotations

import sys
from abc import abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar
from himena.standards.model_meta import DimAxis

import numpy as np

if TYPE_CHECKING:
    from typing import Any, Protocol, TypeAlias, TypeGuard, Self

    import dask.array as da
    import numpy.typing as npt
    import xarray as xr
    import zarr

    _T_contra = TypeVar("_T_contra", contravariant=True)
    Index = int | slice

    class SupportsIndexing(Protocol):
        def __getitem__(self, key: Index | tuple[Index, ...]) -> npt.ArrayLike: ...
        @property
        def shape(self) -> tuple[int, ...]: ...

    class SupportsDunderLT(Protocol[_T_contra]):
        def __lt__(self, other: _T_contra, /) -> bool: ...

    class SupportsDunderGT(Protocol[_T_contra]):
        def __gt__(self, other: _T_contra, /) -> bool: ...

    SupportsRichComparison: TypeAlias = SupportsDunderLT[Any] | SupportsDunderGT[Any]


ArrayT = TypeVar("ArrayT")


class ArrayWrapper(Generic[ArrayT]):
    """Interface for wrapping different array-like data types."""

    def __init__(self, data: ArrayT) -> None:
        self._arr = data

    def __getitem__(self, sl: tuple[Index, ...]) -> Self:
        is_scalar = all(hasattr(s, "__index__") for s in sl)
        if is_scalar:
            return self._arr[sl]
        return self.__class__(self._arr[sl])

    def __setitem__(self, sl: tuple[Index, ...], value: ArrayT) -> None:
        self._arr[sl] = value

    @property
    def arr(self) -> ArrayT:
        return self._arr

    @abstractmethod
    def get_slice(self, sl: tuple[int, ...]) -> np.ndarray:
        """Return a 2D slice of the array as a numpy array."""

    @property
    @abstractmethod
    def dtype(self) -> np.dtype:
        """Return the data type of the array."""

    @property
    @abstractmethod
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the array."""

    def copy(self) -> Self:
        """Return a copy of the array."""
        return self.__class__(self._arr.copy())

    @property
    def size(self) -> int:
        return np.prod(self.shape)

    @property
    def ndim(self) -> int:
        return len(self.shape)

    def model_type(self) -> str:
        typ = type(self._arr)
        return f"{typ.__module__}.{typ.__name__}"

    def infer_axes(self) -> list[DimAxis]:
        """Infer DimAxis objects for this array."""
        return [DimAxis(name=f"axis_{i}") for i in range(self.ndim)]


class XarrayWrapper(ArrayWrapper["xr.DataArray"]):
    """Wrapper for xarray DataArray objects."""

    def get_slice(self, sl: tuple[int, ...]) -> np.ndarray:
        return self._arr[sl].values

    @property
    def dtype(self) -> np.dtype:
        return self._arr.dtype

    @property
    def shape(self) -> tuple[int, ...]:
        return self._arr.shape

    def infer_axes(self) -> list[DimAxis]:
        """Infer DimAxis objects for this array."""

        axes = []
        for name, coord in self._arr.coords.items():
            if is_xarray(coord):
                unit = coord.attrs.get("units", None)
            else:
                unit = None
            axes.append(DimAxis(name=str(name), unit=unit))
        return axes


class ArrayLikeWrapper(ArrayWrapper[ArrayT]):
    """Wrapper for numpy duck array-like objects."""

    def get_slice(self, sl: tuple[int, ...]) -> np.ndarray:
        return self._asarray(self._arr[sl])

    @staticmethod
    def _asarray(data: ArrayT) -> np.ndarray:
        return np.asarray(data)

    @property
    def dtype(self) -> np.dtype:
        return self._arr.dtype

    @property
    def shape(self) -> tuple[int, ...]:
        return self._arr.shape


class DaskWrapper(ArrayLikeWrapper["da.Array"]):
    """Wrapper for dask array objects."""

    def get_slice(self, sl: tuple[int, ...]) -> np.ndarray:
        return self._arr[sl].compute()


class ZarrArrayWrapper(ArrayLikeWrapper["zarr.Array"]):
    """Wrapper for zarr array objects."""

    def __init__(self, data: Any) -> None:
        super().__init__(data)
        self._names = []
        if "_ARRAY_DIMENSIONS" in data.attrs:
            self._names = data.attrs["_ARRAY_DIMENSIONS"]
        else:
            self._names = list(range(data.ndim))


def _see_imported_module(arr: Any, module: str) -> bool:
    typ = type(arr)
    if module not in sys.modules or typ.__module__.split(".")[0] != module:
        return False
    return True


def is_numpy(data: Any) -> TypeGuard[np.ndarray]:
    return isinstance(data, np.ndarray)


def is_dask(data: Any) -> TypeGuard[da.Array]:
    if _see_imported_module(data, "dask"):
        import dask.array as da

        return isinstance(data, da.Array)
    return False


def is_xarray(data: Any) -> TypeGuard[xr.DataArray]:
    if _see_imported_module(data, "xarray"):
        import xarray as xr

        return isinstance(data, xr.DataArray)
    return False


def is_zarr(data: Any) -> TypeGuard[zarr.Array]:
    if _see_imported_module(data, "zarr"):
        import zarr

        return isinstance(data, zarr.Array)
    return False


def wrap_array(arr: Any) -> ArrayWrapper:
    if isinstance(arr, ArrayWrapper):
        return arr
    if is_numpy(arr):
        return ArrayLikeWrapper(arr)
    if is_dask(arr):
        return DaskWrapper(arr)
    if is_xarray(arr):
        return XarrayWrapper(arr)
    if is_zarr(arr):
        return ArrayLikeWrapper(arr)
    return ArrayLikeWrapper(arr)



from __future__ import annotations

from functools import reduce
import operator
from typing import Any, Sequence, TYPE_CHECKING

from app_model.types import KeyBindingRule
from app_model.expressions import BoolOp

from himena._app_model import AppContext as ctx

if TYPE_CHECKING:
    KeyBindingsType = str | KeyBindingRule | Sequence[str] | Sequence[KeyBindingRule]
    PluginConfigType = Any


def expr_and(expr: BoolOp | None, other: BoolOp) -> BoolOp:
    if expr is None:
        return other
    return expr & other


def types_to_expression(types: list[str]) -> BoolOp:
    return reduce(operator.or_, map(type_to_expression, types))


def type_to_expression(typ: str) -> BoolOp:
    subtypes = typ.split(".")
    nsub = len(subtypes)
    out = ctx.active_window_model_type == subtypes[0]
    if nsub >= 2:
        out &= ctx.active_window_model_subtype_1 == subtypes[1]
        if nsub >= 3:
            out &= ctx.active_window_model_subtype_2 == subtypes[2]
            if nsub >= 4:
                out &= ctx.active_window_model_subtype_3 == subtypes[3]
                if nsub >= 5:
                    raise ValueError(f"The maximum number of subtypes are 4, got {typ}")
    return out


def is_model_menu_prefix(menu_id: str) -> bool:
    ids = menu_id.split("/")
    if len(ids) < 3:
        return False
    return (ids[0], ids[1]) == ("", "model_menu")



from __future__ import annotations
from typing import Callable, TYPE_CHECKING

from app_model import Application
import in_n_out
from himena._app_model._command_registry import CommandsRegistry
from himena.types import WidgetDataModel, FutureInfo, WindowRect

if TYPE_CHECKING:
    from concurrent.futures import Future


def get_model_app(name: str) -> HimenaApplication:
    if name in HimenaApplication._instances:
        return HimenaApplication._instances[name]
    return HimenaApplication(name)


class HimenaApplication(Application):
    """The Application class for Himena."""

    def __init__(self, name: str):
        super().__init__(
            name,
            commands_reg_class=CommandsRegistry,
            injection_store_class=HimenaInjectionStore,
            raise_synchronous_exceptions=True,
        )
        self._dynamic_command_ids: set[str] = set()
        self._futures: set[Future] = set()
        self._attributes: dict[str, object] = {}

    @property
    def attributes(self) -> dict[str, object]:
        """A dictionary of attributes for this application."""
        return self._attributes

    @property
    def commands(self) -> CommandsRegistry:
        """The command registry for this application."""
        return super().commands

    def _future_done_callback(self, f: Future) -> None:
        from himena.widgets import current_instance

        self._futures.discard(f)
        if f.cancelled():
            return
        result = f.result()
        if info := FutureInfo.get(f):
            type_hint = info.type_hint
        else:
            type_hint = None
        if isinstance(result, WidgetDataModel) and info is not None:
            if info.track is not None and len(result.workflow) == 0:
                result.workflow = info.track.to_workflow(info.kwargs)
            if result.update_inplace and info.track and info.track.contexts:
                ui = current_instance()
                input_window = ui._window_for_workflow_id(info.track.contexts[0].value)
                input_window.update_model(result)
                input_window._update_model_workflow(result.workflow)
                return None  # no need to process
            elif (top_left := info.top_left) is not None:
                # this clause is used to move the output window to the geometry of
                # preview window or the parametric window.
                _left, _top = top_left
                if info.size is not None:
                    result.window_rect_override = lambda _: WindowRect(
                        _left, _top, *info.size
                    )
                else:
                    result.window_rect_override = lambda size: WindowRect(
                        _left, _top, *size
                    )
        self.injection_store.process(result, type_hint=type_hint)


class HimenaInjectionStore(in_n_out.Store):
    def process(
        self,
        result,
        *,
        type_hint: object | None = None,
        first_processor_only: bool = False,
        raise_exception: bool = True,  # update default
        _funcname: str = "",
    ) -> None:
        super().process(
            result, type_hint=type_hint, first_processor_only=first_processor_only,
            raise_exception=raise_exception, _funcname=_funcname
        )  # fmt: skip

    def inject_processors(self, *args, **kwargs) -> Callable:
        kwargs["raise_exception"] = True
        return super().inject_processors(*args, **kwargs)
